import unittest
import logging
import tests.utils as utils
from tests.utils import compare_json_return_errormessage as error_message
import esgfpid.assistant.datacart

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS


'''
Unit tests for esgfpid.assistant.datacart.

This module needs a coupler to work: It forwards calls to
RabbitMQ to the coupler.

For the tests, we use a real coupler object that has a
mocked RabbitMQ connection., and whose solr communication
is simply switched off (not mocked), as it is never used
for errata operations anyway.
'''
class DataCartTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    '''
    Trivial test of the constructor.
    Only check if instantiation works without errors.
    '''
    def test_init_ok(self):

        # Preparation: Make patched test coupler
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)

        # Run code to be tested:
        assistant = esgfpid.assistant.datacart.DataCartAssistant(
            prefix='foo',
            coupler=testcoupler
        )

        # Check result:
        self.assertIsInstance(assistant, esgfpid.assistant.datacart.DataCartAssistant, 'Constructor fail.')

    '''
    Make sure that the PID of a datacart with just one dataset
    is not the same as the dataset PID itself!
    '''
    def test_if_datacart_pid_same_as_dataset_pid(self):

        # Test variables:
        prefix = 'myprefix'
        drs_id = 'mydrsfoo'
        vers_num = 20160101
        # Dataset id and PID
        dataset_id = esgfpid.utils.concatenate_drs_and_versionnumber(
            drs_id,
            vers_num
        )
        dataset_pid = esgfpid.assistant.publish.create_dataset_handle(
            drs_id = drs_id,
            version_number = vers_num,
            prefix = prefix
        )
        # Data cart content:
        content = {dataset_id : dataset_pid}

        # Preparations: Make an assistant with a patched coupler.
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        assistant = esgfpid.assistant.datacart.DataCartAssistant(
            prefix=prefix,
            coupler=testcoupler
        )

        # Run code to be tested:
        datacart_pid = assistant.make_data_cart_pid(content)

        # Check result
        self.assertNotEqual(dataset_pid, datacart_pid)

    '''
    Test if we can create a data cart PID for several
    datasets that all have PIDs.
    '''
    def test_datacart_for_several_datasets_with_pids(self):

        # Test variables
        content = {'foo':'foo', 'bar':'bar'}
  
        # Preparations: Make an assistant with a patched coupler.
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        assistant = esgfpid.assistant.datacart.DataCartAssistant(
            prefix=PREFIX_NO_HDL,
            coupler=testcoupler
        )

        # Run code to be tested:
        assistant.make_data_cart_pid(content)

        # Check result:
        expected_rabbit_task = {
            "handle": PREFIX_WITH_HDL+'/b597a79e-1dc7-3d3f-b689-75ac5a78167f',
            "operation": "shopping_cart",
            "message_timestamp":"anydate",
            "data_cart_content":{'foo':'foo', 'bar':'bar'},
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.datacart'
        }
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    '''
    Test whether we can create a data cart PID even
    if some datasets have no PID.
    '''
    def test_datacart_some_datasets_have_no_handles(self):

        # Test variablesl
        content = {'foo':'foo', 'bar':None}
  
        # Preparations: Make an assistant with a patched coupler.
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        assistant = esgfpid.assistant.datacart.DataCartAssistant(
            prefix=PREFIX_NO_HDL,
            coupler=testcoupler
        )

        # Run code to be tested:
        assistant.make_data_cart_pid(content)

        # Check result:
        expected_rabbit_task = {
            "handle": PREFIX_WITH_HDL+'/b597a79e-1dc7-3d3f-b689-75ac5a78167f',
            "operation": "shopping_cart",
            "message_timestamp":"anydate",
            "data_cart_content": content,
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.datacart'
        }
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    '''
    Test whether the same data cart PID is created/updated, if
    the same datasets are given, even if dataset PIDs may vary
    (with/without "hdl:", or nonexistent handle strings).
    '''
    def test_datacart_several_times_same_datasets(self):

        # Test variables
        content1 = {'foo':'foo', 'bar':'bar'}
        content2 = {'foo':'foo', 'bar': None}
        content3 = {'foo':'foo', 'bar':'hdl:bar'}
        # Note: In all three cases, the same datasets are passed, so
        # the same data cart PID has to be created/updated.

        # Preparations: Make an assistant with a patched coupler.
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        assistant = esgfpid.assistant.datacart.DataCartAssistant(
            prefix=PREFIX_NO_HDL,
            coupler=testcoupler
        )

        # Run code to be tested:
        pid1 = assistant.make_data_cart_pid(content1)
        pid2 = assistant.make_data_cart_pid(content2)
        pid3 = assistant.make_data_cart_pid(content3)

        # Check result:
        # These are the expected messages:
        expected_handle_all_cases = PREFIX_WITH_HDL+"/b597a79e-1dc7-3d3f-b689-75ac5a78167f"
        expected_rabbit_task1 = {
            "handle": expected_handle_all_cases,
            "operation": "shopping_cart",
            "message_timestamp":"anydate",
            "data_cart_content":content1,
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.datacart'
        }
        expected_rabbit_task2 = {
            "handle": expected_handle_all_cases,
            "operation": "shopping_cart",
            "message_timestamp":"anydate",
            "data_cart_content":content2,
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.datacart'
        }
        expected_rabbit_task3 = {
            "handle": expected_handle_all_cases,
            "operation": "shopping_cart",
            "message_timestamp":"anydate",
            "data_cart_content":content3,
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.datacart'
        }
        # Check if all the messages are correct:
        received_rabbit_task1 = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        received_rabbit_task2 = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        received_rabbit_task3 = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 2)
        same1 = utils.is_json_same(expected_rabbit_task1, received_rabbit_task1)
        same2 = utils.is_json_same(expected_rabbit_task2, received_rabbit_task2)
        same3 = utils.is_json_same(expected_rabbit_task3, received_rabbit_task3)
        self.assertTrue(same1, error_message(expected_rabbit_task1, received_rabbit_task1))
        self.assertTrue(same2, error_message(expected_rabbit_task2, received_rabbit_task2))
        self.assertTrue(same3, error_message(expected_rabbit_task3, received_rabbit_task3))
        # Check if all the handles are the same:
        self.assertTrue(pid1==pid2, 'Pids 1&2 are not the same.')
        self.assertTrue(pid1==pid3, 'Pids 1&3 are not the same.')



'''
This module provides a function that allows to capture
messages printed to stout for checking if the error
messages got printed.

The idea is from Stack Overflow:
http://stackoverflow.com/questions/4219717/how-to-assert-output-with-nosetest-unittest-in-python

'''

from contextlib import contextmanager
import sys
if sys.version[0] == '2':
    from StringIO import StringIO
else:
    from io import StringIO

@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


import logging
import esgfpid
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class Checker(object):

    def __init__(self, **args):

        mandatory_args = ['drs_id', 'data_node', 'version_number', 'coupler']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

        # Needed to retrieve stuff from solr:
        self.__coupler = args['coupler']
        self.__drs_id = args['drs_id']
        self.__version_number = args['version_number']
        self.__data_node = args['data_node']

        # Other         
        self.__list_of_previous_files = None # from solr
        self.__will_run_check = True
        self.__message_why_not = 'No reason specified.'

        # Is solr switched off?
        if self.__coupler.is_solr_switched_off():
            self.__will_run_check = False
        else:
            # query solr for previous file handles:
            self.__data_consistency_check_preparation()


    def can_run_check(self):
        return self.__will_run_check

    ### Preparation ###

    def __data_consistency_check_preparation(self):
        logdebug(LOGGER, 'Asking solr for info for consistency check.')
        self.__retrieve_list_of_previous_files_from_solr()
        logdebug(LOGGER, 'Asking solr for info for consistency check... done.')
        self.__decide_whether_run_check_and_inform()

    def __retrieve_list_of_previous_files_from_solr(self):
        self.__list_of_previous_files = None
        try:
            self.__list_of_previous_files = self.__retrieve_from_solr()
        except (esgfpid.exceptions.SolrSwitchedOff, esgfpid.exceptions.SolrError) as e:
            self.__will_run_check = False
            self.__log_solr_error_occured(e)            

    def __retrieve_from_solr(self):
        files = self.__coupler.retrieve_file_handles_of_same_dataset(
            drs_id=self.__drs_id,
            version_number=self.__version_number,
            data_node=self.__data_node
        )
        return files

    def __decide_whether_run_check_and_inform(self):
        if self.__list_of_previous_files is not None:
            
            if len(self.__list_of_previous_files) > 0:
                self.__will_run_check = True
                self.__log_previously_stored_files_found()

            else:
                self.__will_run_check = False
                self.__log_previously_stored_was_empty()
                
        else:
            self.__will_run_check = False
            self.__log_no_useful_info()

    def __log_solr_error_occured(self, exception):
        msg = ('No dataset integrity test could be run. Message: '+repr(exception))
        self.__message_why_not = msg
        loginfo(LOGGER, msg)

    def __log_previously_stored_files_found(self):
        concat_files = ', '.join(self.__list_of_previous_files)
        type_files = type(self.__list_of_previous_files)
        logdebug(LOGGER, 'Previously published fileset: %s (%s)', concat_files, type_files)
        loginfo(LOGGER, 'Data integrity check will be run after files were specified.')

    def __log_previously_stored_was_empty(self):
        msg = ('Data integrity check: No previous information stored '+
               '(apparently this is a first publication).')
        self.__message_why_not = msg
        loginfo(LOGGER, msg)

    
    def __log_no_useful_info(self):
        msg = ('Solr did not return any useful information, '+
               'no dataset integrity test could be run '+
               '(in case this was not a first publication).')
        self.__message_why_not = msg
        loginfo(LOGGER, msg)


    ### Perform the check ###

    def data_consistency_check(self, list_of_given_files):
        if self.__will_run_check:

            if list_of_given_files is None or len(list_of_given_files)==0:
                raise ValueError('Please provide a list of file handles for the consistency check.')

            return self.__data_consistency_check(list_of_given_files)
        else:
            raise ValueError('Consistency check can not be run. Reason: %s' % self.__message_why_not)

    def __data_consistency_check(self, list_of_given_files):
        logdebug(LOGGER, 'Performing consistency check...')
        list_too_many_files = self.__check_for_superfluous_files(list_of_given_files)
        list_missing_files = self.__check_for_missing_files(list_of_given_files)

        any_inconsistency = len(list_missing_files) > 0 or len(list_too_many_files) > 0
        
        if any_inconsistency:
            self.__log_missing_files_if_any(list_missing_files)
            self.__log_too_many_files_if_any(list_too_many_files)
            logdebug(LOGGER, 'Performing consistency check... done (fail)')
            return False
        else:
            logdebug(LOGGER, 'Performing consistency check... done (success)')
            return True

    def __check_for_superfluous_files(self, list_of_given_files):
        list_too_many_files = []
        for file_handle in list_of_given_files:
            if file_handle not in self.__list_of_previous_files:
                list_too_many_files.append(file_handle)
        return list_too_many_files

    def __check_for_missing_files(self, list_of_given_files):
        list_missing_files = []
        for file_handle in self.__list_of_previous_files:
            if file_handle not in list_of_given_files:
                list_missing_files.append(file_handle)
        return list_missing_files

    def __log_too_many_files_if_any(self, list_too_many_files):
        if len(list_too_many_files)>0:
            list_string = ','.join(list_too_many_files)
            LOGGER.warning('Some superfluous files were added in republication of dataset: %s',
                list_string)

    def __log_missing_files_if_any(self, list_missing_files):
        if len(list_missing_files)>0:
            list_string = ','.join(list_missing_files)
            LOGGER.warning(
                'Some files were missing in republication of dataset: %s',
                list_string)



'''
The ESGF PID system relies on a RabbitMQ instance consisting
of various exchanges and queues. Messages are routed to
these queues using routing keys.

This file contains the various routing keys that are used.

The esgfpid library uses routing keys for topic exchanges,
consisting of four "words" (separated by dots). The first
word is the prefix (but without dots), the second one is
the word "HASH" (placeholder for possibe future load balancing
using hash values), the third is an instruction for the Rabbit,
and the fourth one is a short description of the type of message
being sent.

This way, messages can be routed to different consumers based
on the prefix, they can be prioritized based on their content,
and filtered using some instructions.

'''


'''
This is the Rabbit instruction. It tells the RabbitMQ system
that the message received is a freshly published one, in
opposition to messages that have been parked and are retried
several times.
Please note that the instruction can be extended by the
functions below.
'''
RABBIT_INSTRUCTION = 'fresh'


'''
These are the templates for the routing keys, based on the
message's content.
'''
ROUTING_KEYS_TEMPLATES = dict(
    publi_file = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.publi-file-orig',
    publi_file_rep = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.publi-file-repli',
    publi_ds = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.publi-ds-orig',
    publi_ds_rep = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.publi-ds-repli',
    unpubli_all = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.unpubli-allvers',
    unpubli_one = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.unpubli-onevers',
    err_add = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.errata-add',
    err_rem = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.errata-rem',
    data_cart = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.datacart',
    pre_flight = 'PREFIX.HASH.'+RABBIT_INSTRUCTION+'.preflightcheck'
)

'''
This is where the routing keys will be stored, when the
real prefix has been added.
'''
ROUTING_KEYS = dict()

'''
Helper: Remove all dots from the prefix so that it can
be used for routing keys. Dots in routing keys are word
separators.
'''
def _sanitize_prefix(prefix):
    prefix = prefix
    if 'hdl:' in prefix:
        prefix = prefix.replace('hdl:','')
    return prefix.replace('.', '')

'''
Add the (sanitized) prefix to all routing keys.
'''
def add_prefix_to_routing_keys(prefix):
    sanitized_prefix = _sanitize_prefix(prefix)
    for k,v in ROUTING_KEYS_TEMPLATES.items():
        ROUTING_KEYS[k] = v.replace('PREFIX', sanitized_prefix)




'''
These three keywords may be added to the Rabbit instruction
part of the routing key, in case a message is not sent
via a trusted RabbitMQ node.
Currently, this should not happen, as we raise an exception
if an untrusted node is used.
'''
ROUTING_KEY_INTERFIX_UNSURE_IF_TRUSTED = 'untrusted-unsure'
ROUTING_KEY_INTERFIX_UNTRUSTED_AS_FALLBACK = 'untrusted-fallback'
ROUTING_KEY_INTERFIX_UNTRUSTED = 'untrusted-only'


def adapt_routing_key_for_untrusted(routing_key):
    routing_key = routing_key.replace('fresh', 'fresh-'+ROUTING_KEY_INTERFIX_UNTRUSTED)
    return routing_key

def adapt_routing_key_for_untrusted_fallback(routing_key):
    routing_key = routing_key.replace('fresh', 'fresh-'+ROUTING_KEY_INTERFIX_UNTRUSTED_AS_FALLBACK)
    return routing_key

def adapt_routing_key_for_untrusted_unsure(routing_key):
    routing_key = routing_key.replace('fresh', 'fresh-'+ROUTING_KEY_INTERFIX_UNSURE_IF_TRUSTED)
    return routing_key

'''
The default routing key, in case no key is included
in the message. This can only happen if there is some
severe error in the library. Or in testing situations.
'''
RABBIT_DEFAULT_ROUTING_KEY='fallback.fallback.'+RABBIT_INSTRUCTION+'.fallback'

'''
The default routing key, in case a message comes back
from the RabbitMQ as "unroutable". This can only happen
if there is a severe problem with the RabbitMQ federation
(i.e. the exchanges and bindings are not correct, the
alternate-exchange did not catch the message).
'''
RABBIT_EMERGENCY_ROUTING_KEY='UNROUTABLE.UNROUTABLE.'+RABBIT_INSTRUCTION+'.UNROUTABLE'




import unittest
import logging
import sys
if sys.version[0] == '2':
    import Queue as queue
else:
    import queue as queue
import pika
import esgfpid.rabbit.asynchronous.thread_feeder
from esgfpid.rabbit.asynchronous.exceptions import OperationNotAllowed

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

class ThreadFeederTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def make_feeder(self, error=None):

        thread = TESTHELPERS.get_thread_mock2(error)
        statemachine = esgfpid.rabbit.asynchronous.thread_statemachine.StateMachine()
        nodemanager = TESTHELPERS.get_nodemanager()

        feeder = esgfpid.rabbit.asynchronous.thread_feeder.RabbitFeeder(
            thread,
            statemachine,
            nodemanager)

        statemachine.set_to_available()
        nodemanager.set_next_host() # otherwise, we cannot call its method inside the publish method
        return feeder, thread

    # Tests

    #
    # Sending messages
    #

    def test_send_message_ok(self):

        # Preparation:
        msg = "{'foo':'bar'}"
        feeder, thread = self.make_feeder()
        thread.messages.append(msg)

        # Run code to be tested:
        feeder.publish_message()

        # Check result:
        # Publish was called:
        thread._channel.basic_publish.assert_called_once()
        # Message not waiting in queue anymore:
        self.assertNotIn(msg, thread.messages)

    def test_routing_key_ok(self):

        # Preparation:
        msg = '{"foo":"bar", "ROUTING_KEY":"myprefix.HASH.fresh.mydescription"}'
        feeder, thread = self.make_feeder()
        thread.messages.append(msg)

        # Run code to be tested:
        feeder.publish_message()

        # Check result:
        # Publish was called:
        thread._channel.basic_publish.assert_called_once()
        # Message not waiting in queue anymore:
        self.assertNotIn(msg, thread.messages)
        # Correct routing key:
        rk = thread._channel.basic_publish.call_args[1]["routing_key"]
        self.assertEqual(rk, 'myprefix.HASH.fresh.mydescription')

    def test_routing_key_untrusted_ok(self):

        # Preparation:
        msg = '{"foo":"bar", "ROUTING_KEY":"myprefix.HASH.fresh.mydescription"}'
        feeder, thread = self.make_feeder()
        feeder.nodemanager._NodeManager__current_node['is_open']=True
        thread.messages.append(msg)

        # Run code to be tested:
        feeder.publish_message()

        # Check result:
        # Publish was called:
        thread._channel.basic_publish.assert_called_once()
        # Message not waiting in queue anymore:
        self.assertNotIn(msg, thread.messages)
        # Correct routing key:
        rk = thread._channel.basic_publish.call_args[1]["routing_key"]
        self.assertEqual(rk, 'myprefix.HASH.fresh-untrusted-fallback.mydescription')

    def test_routing_key_untrusted_ok(self):

        # Preparation:
        msg = '{"foo":"bar", "ROUTING_KEY":"myprefix.HASH.fresh.mydescription"}'
        feeder, thread = self.make_feeder()
        feeder.nodemanager._NodeManager__current_node['is_open']=True
        feeder.nodemanager._NodeManager__has_trusted=False
        thread.messages.append(msg)

        # Run code to be tested:
        feeder.publish_message()

        # Check result:
        # Publish was called:
        thread._channel.basic_publish.assert_called_once()
        # Message not waiting in queue anymore:
        self.assertNotIn(msg, thread.messages)
        # Correct routing key:
        rk = thread._channel.basic_publish.call_args[1]["routing_key"]
        self.assertEqual(rk, 'myprefix.HASH.fresh-untrusted-only.mydescription')

    def test_send_message_empty(self):

        # Preparation:
        feeder, thread = self.make_feeder()

        # Run code to be tested:
        feeder.publish_message()
        feeder.publish_message()

        # Check result:
        # Publish was not called:
        thread._channel.basic_publish.assert_not_called()
  

    def test_send_message_error(self):

        # Preparation:
        msg = "{'foo':'bar'}"
        feeder, thread = self.make_feeder(error=pika.exceptions.ChannelClosed(reply_code=0, reply_text='Channel Closed'))
        thread.messages.append(msg)

        # Run code to be tested:
        feeder.publish_message()

        # Check result:
        # Publish was tried:
        thread._channel.basic_publish.assert_called_once()
        # Message was put back to queue:
        self.assertIn(msg, thread.messages)
        self.assertIn(msg, thread.put_back)

    def test_send_message_NOT_STARTED_YET(self):

        # Preparation:
        msg = "{'foo':'bar'}"
        feeder, thread = self.make_feeder()
        thread.messages.append(msg)
        feeder.statemachine._StateMachine__state = 0
        
        # Run code to be tested:
        feeder.publish_message()

        # Check result:
        self.assertTrue(feeder.statemachine.is_NOT_STARTED_YET())
        self.assertIn(msg, thread.messages)
        thread._channel.basic_publish.assert_not_called()

    def test_send_message_PERMANENTLY_UNAVAILABLE_1(self):

        # Preparation:
        msg = "{'foo':'bar'}"
        feeder, thread = self.make_feeder()
        thread.messages.append(msg)
        feeder.statemachine.set_to_permanently_unavailable()
        feeder.statemachine.set_detail_closed_by_publisher()

        # Run code to be tested:
        feeder.publish_message()

        # Check result:
        self.assertTrue(feeder.statemachine.is_PERMANENTLY_UNAVAILABLE())
        self.assertIn(msg, thread.messages)
        thread._channel.basic_publish.assert_not_called()

    def test_send_message_PERMANENTLY_UNAVAILABLE_2(self):

        # Preparation:
        msg = "{'foo':'bar'}"
        feeder, thread = self.make_feeder()
        thread.messages.append(msg)
        feeder.statemachine.set_to_permanently_unavailable()
        feeder.statemachine.detail_could_not_connect = True

        # Run code to be tested:
        feeder.publish_message()

        # Check result:
        self.assertTrue(feeder.statemachine.is_PERMANENTLY_UNAVAILABLE())
        self.assertIn(msg, thread.messages)
        thread._channel.basic_publish.assert_not_called()

    def test_send_message_WAITING_TO_BE_AVAILABLE(self):

        # Preparation:
        msg = "{'foo':'bar'}"
        feeder, thread = self.make_feeder()
        thread.messages.append(msg)
        feeder.statemachine.set_to_waiting_to_be_available()
        
        # Run code to be tested:
        feeder.publish_message()

        # Check result:
        self.assertIn(msg, thread.messages)
        self.assertTrue(feeder.statemachine.is_WAITING_TO_BE_AVAILABLE())
        thread._channel.basic_publish.assert_not_called()

    def test_reset_delivery_number(self):

        # Preparation:
        msg = "{'foo':'bar'}"
        feeder, thread = self.make_feeder()
        thread.messages.append(msg)
        thread.messages.append(msg)
        thread.messages.append(msg)

        # Pre-Check:
        self.assertEqual(feeder._RabbitFeeder__delivery_number, 1)

        # Increase delivery number:
        feeder.publish_message()
        feeder.publish_message()
        feeder.publish_message()
        thread._channel.basic_publish.assert_called()

        # Check if it was increased:
        self.assertEqual(feeder._RabbitFeeder__delivery_number, 4)

        # Run code to be tested:
        feeder.reset_delivery_number()

        # Check if it was reset:
        self.assertEqual(feeder._RabbitFeeder__delivery_number, 1)






import unittest
import logging

import esgfpid.defaults
import esgfpid.rabbit.rabbitutils as rutils

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

LOGGER_TO_PASS = logging.getLogger('utils')
LOGGER_TO_PASS.addHandler(logging.NullHandler())


'''
Unit tests for esgfpid.rabbit.rabbitutils.

This module does not need any other module to
function, so we don't need to use or to mock any
other objects. 
'''
class RabbitUtilsTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)


    def tearDown(self):
        LOGGER.info('#############################')

    #
    # Test getting routing key and string message
    #

    def test_get_message_and_routing_key_string_ok(self):
        
        # Test variables:
        passed_message = '{"bla":"foo", "ROUTING_KEY":"roukey"}'
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)

        # Check result:
        received_message = received_message.replace("'", '"')
        expected_message = str(passed_message).replace("'", '"')
        self.assertEqual(received_key, 'roukey', 'Wrong routing key: %s' % received_key)
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))

    def test_get_message_and_routing_key_string_singlequtoes_ok(self):
        
        # Test variables:
        passed_message = "{'bla':'foo', 'ROUTING_KEY':'roukey'}"
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)

        # Check result:
        received_message = received_message.replace("'", '"')
        expected_message = str(passed_message).replace("'", '"')
        self.assertEqual(received_key, 'roukey', 'Wrong routing key: %s' % received_key)
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))

    def test_get_message_and_routing_key_string_no_key(self):
        
        # Test variables:
        passed_message = '{"bla":"foo", "no_key":"no_key"}'
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)

        # Check result:
        received_message = received_message.replace("'", '"')
        expected_message = str(passed_message).replace("'", '"')
        expected_key = esgfpid.utils.RABBIT_DEFAULT_ROUTING_KEY
        self.assertEqual(received_key, expected_key, 'Wrong routing key: %s' % received_key)
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))

    def test_get_message_and_routing_key_none_error(self):
        
        # Test variables:
        passed_message = None
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        with self.assertRaises(ValueError):
            received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)

    def test_get_message_and_routing_key_characters(self):
        
        # Test variables:
        passed_message = 'foo'
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)

        # Check result:
        received_message = received_message.replace("'", '"')
        expected_message = str(passed_message).replace("'", '"')
        expected_key = esgfpid.utils.RABBIT_DEFAULT_ROUTING_KEY
        self.assertEqual(received_key, expected_key, 'Wrong routing key: %s' % received_key)
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))

    def test_get_message_and_routing_key_json_ok(self):
        
        # Test variables:
        passed_message = {"bla":"foo", "ROUTING_KEY":"roukey"}
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)

        # Check result:
        received_message = received_message.replace("'", '"')
        expected_message = str(passed_message).replace("'", '"')
        self.assertEqual(received_key, 'roukey', 'Wrong routing key: %s' % received_key)
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))

    def test_get_message_and_routing_key_json_no_rk(self):
        
        # Test variables:
        passed_message = {"bla":"foo", "no_key":"no_key"}
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)

        # Check result:
        received_message = received_message.replace("'", '"')
        expected_message = str(passed_message).replace("'", '"')
        expected_key = esgfpid.utils.RABBIT_DEFAULT_ROUTING_KEY
        self.assertEqual(received_key, expected_key, 'Wrong routing key: %s' % received_key)
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))

    def test_get_message_and_routing_key_list(self):
        
        # Test variables:
        passed_message = [345]
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)

        # Check result:
        received_message = received_message.replace("'", '"')
        expected_message = str(passed_message).replace("'", '"')
        expected_key = esgfpid.utils.RABBIT_DEFAULT_ROUTING_KEY
        self.assertEqual(received_key, expected_key, 'Wrong routing key: %s' % received_key)
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))

    def test_get_message_and_routing_key_int(self):
        
        # Test variables:
        passed_message = 345
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)

        # Check result:
        received_message = received_message.replace("'", '"')
        expected_message = str(passed_message).replace("'", '"')
        expected_key = esgfpid.utils.RABBIT_DEFAULT_ROUTING_KEY
        self.assertEqual(received_key, expected_key, 'Wrong routing key: %s' % received_key)
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))

    def test_get_message_and_routing_key_bogus(self):
        
        # Test variables:
        def foo(): print('yeah')
        passed_message = foo
        LOGGER.info('Message: %s' % passed_message)

        # Run code to be checked:
        with self.assertRaises(ValueError):
            received_key, received_message = rutils.get_routing_key_and_string_message_from_message_if_possible(passed_message)






import esgfpid
import logging
import datetime

'''
A script to test all tasks of the esgfpid library.

This only tests the most basic standard cases.

Please checkout the logs to verify the correct
execution of all functionality. As the module
has two threads and some parallel, asynchronous
execution, errors tend not to show in this script.

'''

# Please fill in these values:
PREFIX = '21.14100' # CMIP6
RABBIT_URL = 'foo.rabbit.bar'
RABBIT_EXCHANGE = 'foo-exchangename'
RABBIT_USER = 'johndoe'
RABBIT_PW = 'insertme'
LOGLEVEL = logging.DEBUG
LOGLEVEL = logging.INFO
RABBIT_USER_OPEN = 'foo-user'
RABBIT_URL_OPEN = 'foo.open.bar'

# For publication:
DATA_NODE1 = 'foo.bar.test'
DATA_NODE2 = 'foo.hello.test'
THREDDS_PATH = '/foo/bar/baz/'

# 
TEST_DRS = 'test.drs.id1'
TEST_VERSION1 = 20001111
TEST_VERSION2 = 20201111

IS_TEST = False

def init_logging():
    path = 'logs'
    logdate = datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
    filename = path+'/test_log_'+logdate+'_'+logging.getLevelName(LOGLEVEL)+'.log'
    esgfpid.utils.ensure_directory_exists(path)
    myformat = '%(asctime)-15s %(levelname)-5s: %(name)s [%(threadName)s]: %(message)s'
    logging.basicConfig(level=LOGLEVEL, filename=filename, filemode='w', format=myformat)
    pikalogger = logging.getLogger('pika')
    pikalogger.setLevel(logging.WARNING)
    print('Logging to file ".%s"' % filename)

def init_connector(data_node):
    trusted_node1 = {
        'user':RABBIT_USER,
        'password':RABBIT_PW,
        'url':RABBIT_URL,
        'priority':1
    }
    trusted_node2 = {
        'user':RABBIT_USER,
        'password':RABBIT_PW,
        'url':RABBIT_URL,
        'priority':2
    }
    open_node = {
        'user':RABBIT_USER_OPEN,
        'url':RABBIT_URL_OPEN
    }
    connector = esgfpid.Connector(
        handle_prefix=PREFIX,
        messaging_service_credentials=[trusted_node1,trusted_node2,open_node],
        messaging_service_exchange_name=RABBIT_EXCHANGE,
        data_node=data_node,
        thredds_service_path=THREDDS_PATH,
        test_publication=IS_TEST
    )
    connector.start_messaging_thread()
    return connector

def publish_dataset(connector, version_number):
    wizard = connector.create_publication_assistant(
        drs_id=TEST_DRS,
        version_number=version_number,
        is_replica=False    
    )
    ds_handle = wizard.get_dataset_handle()
    file_handle1 = PREFIX+'/sdkfjskfjsidlfj'
    file_handle2 = PREFIX+'/jglsrirlsjjlekdfsl'
    wizard.add_file(
        file_name='temperature.nc',
        file_handle=file_handle1,
        file_size=1000,
        checksum='abc123check',
        publish_path='my/path/file.nc',
        checksum_type='SHA256',
        file_version='f1'
    )
    wizard.add_file(
        file_name='temperature.nc',
        file_handle=file_handle2,
        file_size=1000,
        checksum='abc123check',
        publish_path='my/path/differentfile.nc',
        checksum_type='SHA256',
        file_version='f1'
    )
    wizard.dataset_publication_finished()

    print('Publication done:')
    print('hdl.handle.net/%s?noredirect' % ds_handle)
    print('hdl.handle.net/%s?noredirect' % file_handle1)
    print('hdl.handle.net/%s?noredirect' % file_handle2)

def add_and_remove_errata(connector, drs_id, version_number):
    connector.add_errata_ids(
        drs_id=drs_id,
        version_number=version_number,
        errata_ids=['abc','def','ghi']
    )
    connector.remove_errata_ids(
        drs_id=drs_id,
        version_number=version_number,
        errata_ids=['def']
    )

def shopping_cart(connector):
    pid = connector.create_shopping_cart_pid({'foo':'hdl:123/345','bar':None,'baz':'hdl:123/678'})
    print('Shopping cart:')
    print('hdl.handle.net/%s?noredirect' % pid)

def unpublish_one_version(connector, drs_id, version_number):
    connector.unpublish_one_version(
        drs_id=drs_id,
        version_number=version_number
    )

def unpublish_all_versions(connector, drs_id):
    connector.unpublish_all_versions(
        drs_id=drs_id
    )

def cleanup(connector):
    print('Waiting for pending messages and stopping thread...')
    connector.finish_messaging_thread()
    print('Script finished.')
    print('(It may take some minutes until these PIDs are actually created).')


# Now run:
init_logging()

# Publish and unpublish at one data node:
connector1 = init_connector(DATA_NODE1)
publish_dataset(connector1, TEST_VERSION1)
unpublish_all_versions(connector1, TEST_DRS)

# Publish and unpublish at another data node:
connector2 = init_connector(DATA_NODE2)
import time
time.sleep(3)
publish_dataset(connector2, TEST_VERSION1)
unpublish_one_version(connector2, TEST_DRS, TEST_VERSION1)
publish_dataset(connector2, TEST_VERSION2)
add_and_remove_errata(connector2, TEST_DRS, TEST_VERSION1)

# Shopping cart
shopping_cart(connector2)

# Cleanup
cleanup(connector1)
cleanup(connector2)

print('Done!')


import unittest
import logging
import copy
import tests.utils as utils
from tests.utils import compare_json_return_errormessage as error_message

import esgfpid.exceptions
import esgfpid.assistant.messages as messages

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *

class MessageCreationTestcase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)
        esgfpid.utils.routingkeys.add_prefix_to_routing_keys(PREFIX_WITH_HDL)

    def tearDown(self):
        LOGGER.info('#############################')

    def __get_args_dict_file(self):
        return dict(
            file_handle='123/456',
            is_replica='mightbe',
            file_size='333bytes',
            file_name='filey.nc',
            checksum='xyz',
            data_url='myurl.de',
            data_node='dkrz.de',
            parent_dataset='abc/def',
            timestamp='todayish',
            checksum_type='SHA999',
            file_version='v7'
        )

    def __get_args_dict_dataset(self):
        return dict(
            dataset_handle='abc/def',
            is_replica='mightbe',
            drs_id='drs/id/foo',
            version_number='201699',
            list_of_files='listof_files',
            data_node='dkrz.de',
            timestamp='todayish'
        )

    ### Actual test cases: ###

    def test_make_message_publication_file_ok(self):

        # Test variables
        args_dict = self.__get_args_dict_file()
  
        # Run code to be tested:
        received_message = messages.publish_file(**args_dict)

        # Check result:
        expected = copy.deepcopy(args_dict)
        expected['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-file-orig'
        expected['aggregation_level'] = 'file'
        expected['operation'] = 'publish'
        # Rename some:
        expected['handle'] = expected['file_handle']
        del expected['file_handle']
        expected['message_timestamp'] = expected['timestamp']
        del expected['timestamp']

        same = utils.is_json_same(expected, received_message)
        self.assertTrue(same, error_message(expected, received_message))

    def test_make_message_publication_file_replica_ok(self):

        # Test variables
        args_dict = self.__get_args_dict_file()
        args_dict['is_replica'] = True
  
        # Run code to be tested:
        received_message = messages.publish_file(**args_dict)

        # Check result:
        expected = copy.deepcopy(args_dict)
        expected['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-file-repli'
        expected['aggregation_level'] = 'file'
        expected['operation'] = 'publish'
        # Rename some:
        expected['handle'] = expected['file_handle']
        del expected['file_handle']
        expected['message_timestamp'] = expected['timestamp']
        del expected['timestamp']

        same = utils.is_json_same(expected, received_message)
        self.assertTrue(same, error_message(expected, received_message))

    def test_make_message_publication_dataset_ok(self):

        # Test variables
        args_dict = self.__get_args_dict_dataset()
  
        # Run code to be tested:
        received_message = messages.publish_dataset(**args_dict)

        # Check result:
        expected = copy.deepcopy(args_dict)
        expected['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-ds-orig'
        expected['aggregation_level'] = 'dataset'
        expected['operation'] = 'publish'
        # Rename some:
        expected['handle'] = expected['dataset_handle']
        del expected['dataset_handle']
        expected['message_timestamp'] = expected['timestamp']
        del expected['timestamp']
        expected['files'] = expected['list_of_files']
        del expected['list_of_files']

        same = utils.is_json_same(expected, received_message)
        self.assertTrue(same, error_message(expected, received_message))


    def test_make_message_publication_dataset_replica_ok(self):

        # Test variables
        args_dict = self.__get_args_dict_dataset()
        args_dict['is_replica'] = True
  
        # Run code to be tested:
        received_message = messages.publish_dataset(**args_dict)

        # Check result:
        expected = copy.deepcopy(args_dict)
        expected['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-ds-repli'
        expected['aggregation_level'] = 'dataset'
        expected['operation'] = 'publish'
        # Rename some:
        expected['handle'] = expected['dataset_handle']
        del expected['dataset_handle']
        expected['message_timestamp'] = expected['timestamp']
        del expected['timestamp']
        expected['files'] = expected['list_of_files']
        del expected['list_of_files']

        same = utils.is_json_same(expected, received_message)
        self.assertTrue(same, error_message(expected, received_message))


    def test_make_message_publication_dataset_missing_args(self):

        # Test variables
        args_dict = self.__get_args_dict_dataset()
  
        # Run code to be tested:
        for k,v in args_dict.items():
            args = copy.deepcopy(args_dict)
            LOGGER.debug('Deleting %s and trying...', k)
            del args[k]
            with self.assertRaises(esgfpid.exceptions.ArgumentError):
                received_message = messages.publish_dataset(**args)


    def test_make_message_unpublish_allversions_consumer_must_find_versions_ok(self):

        # Test variables
        args_dict = dict(
            drs_id = 'abc',
            data_node = 'dkrz.de',
            timestamp = 'todayish'
        )

        # Run code to be tested:
        received_message = messages.unpublish_allversions_consumer_must_find_versions(**args_dict)

        # Check result:
        expected = copy.deepcopy(args_dict)
        expected['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-allvers'
        expected['aggregation_level'] = 'dataset'
        expected['operation'] = 'unpublish_all_versions'
        # Rename some:
        expected['message_timestamp'] =  expected['timestamp']
        del expected['timestamp']

        same = utils.is_json_same(expected, received_message)
        self.assertTrue(same, error_message(expected, received_message))


    def test_make_message_unpublish_one_version_ok(self):

        # Test variables
        args_dict = dict(
            dataset_handle = 'abc/def',
            data_node = 'dkrz.de',
            timestamp = 'todayish',
            drs_id = 'mytest'
        )

        # Run code to be tested:
        received_message = messages.unpublish_one_version(**args_dict)

        # Check result:
        expected = copy.deepcopy(args_dict)
        expected['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-onevers'
        expected['aggregation_level'] = 'dataset'
        expected['operation'] = 'unpublish_one_version'
        # Rename some:
        expected['handle'] = expected['dataset_handle']
        del expected['dataset_handle']
        expected['message_timestamp'] = expected['timestamp']
        del expected['timestamp']

        same = utils.is_json_same(expected, received_message)
        self.assertTrue(same, error_message(expected, received_message))


    def test_make_message_add_errata_ids_ok(self):

        # Test variables
        args_dict = dict(
            dataset_handle = 'abc/def',
            errata_ids = 'dkrz.de',
            timestamp = 'todayish',
            drs_id = 'drs_foo',
            version_number = 'vers_foo'
        )

        # Run code to be tested:
        received_message = messages.add_errata_ids_message(**args_dict)

        # Check result:
        expected = copy.deepcopy(args_dict)
        expected['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.errata-add'
        expected['operation'] = 'add_errata_ids'
        expected['drs_id'] = 'drs_foo'
        expected['version_number'] = 'vers_foo'
        # Rename some:
        expected['handle'] = expected['dataset_handle']
        del expected['dataset_handle']
        expected['message_timestamp'] = expected['timestamp']
        del expected['timestamp']

        same = utils.is_json_same(expected, received_message)
        self.assertTrue(same, error_message(expected, received_message))

    def test_make_message_remove_errata_ids_ok(self):

        # Test variables
        args_dict = dict(
            dataset_handle = 'abc/def',
            errata_ids = 'dkrz.de',
            timestamp = 'todayish',
            drs_id = 'drs_foo',
            version_number = 'vers_foo'
        )

        # Run code to be tested:
        received_message = messages.remove_errata_ids_message(**args_dict)

        # Check result:
        expected = copy.deepcopy(args_dict)
        expected['ROUTING_KEY'] = PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.errata-rem'
        expected['operation'] = 'remove_errata_ids'
        expected['drs_id'] = 'drs_foo'
        expected['version_number'] = 'vers_foo'
        # Rename some:
        expected['handle'] = expected['dataset_handle']
        del expected['dataset_handle']
        expected['message_timestamp'] = expected['timestamp']
        del expected['timestamp']

        same = utils.is_json_same(expected, received_message)
        self.assertTrue(same, error_message(expected, received_message))


import pika.exceptions
import mock


class MockChannel(object):

    def __init__(self):
        self.is_open = True # This mocks the original API
        self.messages = []
        self.routing_keys = []
        self.publish_counter = 0 # is incremented at every publish
        self.num_failures = 0 # Can be set before mock is called
        self.num_unroutables = 0 # Can be set before mock is called
        self.success_counter = 0
        self.channel_number = 1 # This mocks the original API

    def confirm_delivery(self, *args, **kwargs): # This mocks the original API
        pass

    def basic_publish(self, *args, **kwargs): # This mocks the original API
        self.publish_counter += 1
        self.messages.append(kwargs['body'])
        self.routing_keys.append(kwargs['routing_key'])
        if self.publish_counter <= self.num_unroutables:
            raise pika.exceptions.UnroutableError('Unroutableeee!')
        elif self.publish_counter <= self.num_failures:
            return False
        else:
            self.success_counter += 1
            return True

    def exchange_declare(self, *args, **kwargs): # This mocks the original API
        pass

    def queue_declare(self, *args, **kwargs): # This mocks the original API
        pass

    def queue_bind(self, *args, **kwargs): # This mocks the original API
        pass

    def add_on_close_callback(self, *args, **kwargs): # This mocks the original API
        pass

    def add_on_return_callback(self, *args, **kwargs): # This mocks the original API
        pass

class MockPikaBlockingConnection(object):

    def __init__(self, params):
        self.is_open = True # This mocks the original API
        self.host = params.host # This mocks the original API
        self.raise_channel_closed = False

        #if params.host == 'please.raise.auth.error':
        #    raise pika.exceptions.ProbableAuthenticationError()
        #elif params.host == 'please.raise.connection.error':
        #    raise pika.exceptions.AMQPConnectionError()
        #else:
        #    pass

    def channel(self): # This mocks the original API
        #return MockChannel()
        if self.raise_channel_closed:
            raise pika.exceptions.ChannelClosed(reply_code=0, reply_text='Channel Closed')
        else:
            return MockChannel()

    def close(self): # This mocks the original API
        self.is_open = False

    def process_data_events(self): # This mocks the original API
        pass

class MockPikaSelectConnection(object):

    def __init__(self):
        self.is_open = True
        self.is_closed = False
        self.is_closing = False # for simplicity, leave always false
        self.ioloop = mock.MagicMock()
        self.__channel = MockChannel()

    def channel(self, on_open_callback): # This mocks the original API
        return self.__channel

    def close(self, *args, **kwargs): # This mocks the original API
        self.is_open = False
        self.is_closed = True

    #def add_timeout(self, seconds, to_be_called):
    #    to_be_called()

    def add_on_close_callback(self, callback):
        pass




import esgfpid.defaults
import copy

#
# Logging helpers
#

def logtrace(logger, msg, *args, **kwargs):
    '''
    If esgfpid.defaults.LOG_TRACE_TO_DEBUG, messages are treated
    like debug messages (with an added [trace]).
    Otherwise, they are ignored.
    '''
    if esgfpid.defaults.LOG_TRACE_TO_DEBUG:
        logdebug(logger, '[trace] %s' % msg, *args, **kwargs)
    else:
        pass

def logdebug(logger, msg, *args, **kwargs):
    '''
    Logs messages as DEBUG,
    unless show=True and esgfpid.defaults.LOG_SHOW_TO_INFO=True,
    (then it logs messages as INFO).
    '''
    if esgfpid.defaults.LOG_DEBUG_TO_INFO:
        logger.info('DEBUG %s ' % msg, *args, **kwargs)
    else:
        logger.debug(msg, *args, **kwargs)

def loginfo(logger, msg, *args, **kwargs):
    '''
    Logs messages as INFO,
    unless esgfpid.defaults.LOG_INFO_TO_DEBUG,
    (then it logs messages as DEBUG).
    '''
    if esgfpid.defaults.LOG_INFO_TO_DEBUG:
        logger.debug(msg, *args, **kwargs)
    else:
        logger.info(msg, *args, **kwargs)


def logwarn(logger, msg, *args, **kwargs):
    logger.warn(msg, *args, **kwargs)

def logerror(logger, msg, *args, **kwargs):
    logger.error(msg, *args, **kwargs)

def log_every_x_times(logger, counter, x, msg, *args, **kwargs):
    '''
    Works like logdebug, but only prints first and
    and every xth message.
    '''
    if counter==1 or counter % x == 0:
        #msg = msg + (' (counter %i)' % counter)
        logdebug(logger, msg, *args, **kwargs)

def make_logsafe(list_or_dict, forbidden='password'):
    if not forbidden in str(list_or_dict):
        return list_or_dict

    # don't modify the original:
    logsafe = copy.deepcopy(list_or_dict)

    # Dict (only first level):
    # TODO: Recursively check deeper levels, or count occurrences of 
    # forbidden word to make sure you got them all.
    try:
        logsafe[forbidden] = logsafe[forbidden][0:3]+'...'
    except KeyError:
        pass
    except TypeError:

        # List of dicts (only first level):
        for item in logsafe:
            try:
                item[forbidden] = item[forbidden][0:3]+'...'
            except KeyError:
                pass

    return logsafe




import unittest
import mock
import logging
import esgfpid.rabbit.asynchronous.thread_shutter
from esgfpid.rabbit.asynchronous.exceptions import OperationNotAllowed

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
import resources.TESTVALUES as TESTHELPERS

class ThreadShutterTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def make_shutter(self):

        # Mocked:
        thread = TESTHELPERS.get_thread_mock4()
        statemachine = esgfpid.rabbit.asynchronous.thread_statemachine.StateMachine()
        statemachine._StateMachine__state = statemachine._StateMachine__IS_AVAILABLE

        shutter = esgfpid.rabbit.asynchronous.thread_shutter.ShutDowner(
            thread,
            statemachine)
        return shutter

    #
    # Force finish
    #

    '''
    Simple force-close.
    No messages are left.
    '''
    def test_force_finish_no_leftovers_ok(self):

        # Preparation:
        shutter = self.make_shutter()

        # Run code to be tested:
        shutter.force_finish()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__FORCE_FINISHED
        self.assertEqual(state, exp)
        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 0)
        self.assertEqual(shutter.thread.get_num_unpublished(), 0)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()

    '''
    Force-close when some messages are still unpublished/unconfirmed.
    They stay unpublished/unconfirmed.
    '''
    def test_force_finish_leftovers_ok(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread.num_unconfirmed = 4
        shutter.thread.num_unpublished = 8

        # Run code to be tested:
        shutter.force_finish()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__FORCE_FINISHED
        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 4)
        self.assertEqual(shutter.thread.get_num_unpublished(), 8)
        self.assertFalse(shutter.thread.add_event_publish_message.called)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()

    '''
    Simple force-close.
    No messages are left.
    '''
    def test_force_finish_was_closed(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread._connection.is_closed = True
        shutter.thread._connection.is_open = False

        # Run code to be tested:
        shutter.force_finish()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__FORCE_FINISHED
        self.assertEqual(state, exp)
        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 0)
        self.assertEqual(shutter.thread.get_num_unpublished(), 0)
        # Check if ioloop was closed:
        shutter.thread.make_permanently_closed_by_user.assert_called_with()

    #
    # Gentle finish
    #

    def test_gently_finish_no_leftovers_ok(self):

        # Preparation:
        shutter = self.make_shutter()

        # Run code to be tested:
        shutter.finish_gently()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__PERMANENTLY_UNAVAILABLE
        self.assertEqual(state, exp)
        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 0)
        self.assertEqual(shutter.thread.get_num_unpublished(), 0)
        self.assertFalse(shutter.thread.add_event_publish_message.called)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()
    
    def test_gently_finish_with_unpublished_messages_ok(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread.num_unpublished = 4

        # Run code to be tested:
        shutter.finish_gently()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__PERMANENTLY_UNAVAILABLE
        self.assertEqual(state, exp)

        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 0)
        self.assertEqual(shutter.thread.get_num_unpublished(), 0)
        self.assertTrue(shutter.thread.add_event_publish_message.called)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()

    def test_gently_finish_with_unconfirmed_messages_ok(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread.num_unconfirmed = 4
        shutter.thread.fake_confirms = True
        shutter.thread.confirm_rate = 0.22

        # Run code to be tested:
        shutter.finish_gently()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__PERMANENTLY_UNAVAILABLE
        self.assertEqual(state, exp)

        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 0)
        self.assertEqual(shutter.thread.get_num_unpublished(), 0)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()

    def test_gently_finish_wait_too_long(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread.num_unconfirmed = 100
        shutter.thread.fake_confirms = True
        shutter.thread.confirm_rate = 0.22

        # Run code to be tested:
        shutter.finish_gently()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__FORCE_FINISHED
        self.assertEqual(state, exp)

        # Check leftovers:
        self.assertTrue(shutter.thread.get_num_unconfirmed()>0)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()

    def test_gently_finish_with_leftovers_not_progressing_1(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread.num_unpublished = 100
        shutter.thread.num_unconfirmed = 100
        shutter.thread.fake_confirms = False
        shutter.statemachine.set_to_permanently_unavailable()
        shutter.statemachine.detail_could_not_connect = True

        # Run code to be tested:
        shutter.finish_gently()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__FORCE_FINISHED
        self.assertEqual(state, exp)

        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 100)
        self.assertEqual(shutter.thread.get_num_unpublished(), 100)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()

    '''
    Same as before, but different reason.
    '''
    def test_gently_finish_with_leftovers_not_progressing_2(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread.num_unpublished = 100
        shutter.thread.num_unconfirmed = 100
        shutter.thread.fake_confirms = False
        shutter.statemachine.set_to_permanently_unavailable()
        shutter.statemachine.detail_authentication_exception = True

        # Run code to be tested:
        shutter.finish_gently()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__FORCE_FINISHED
        self.assertEqual(state, exp)

        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 100)
        self.assertEqual(shutter.thread.get_num_unpublished(), 100)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()

    '''
    Same as before, but different reason.
    '''
    def test_gently_finish_with_leftovers_not_progressing_3(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread.num_unpublished = 100
        shutter.thread.num_unconfirmed = 100
        shutter.thread.fake_confirms = False
        shutter.statemachine.set_to_permanently_unavailable()
        shutter.statemachine.set_detail_closed_by_publisher()

        # Run code to be tested:
        shutter.finish_gently()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__FORCE_FINISHED
        self.assertEqual(state, exp)

        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 100)
        self.assertEqual(shutter.thread.get_num_unpublished(), 100)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()

    def test_not_continuing(self):

        # Preparation:
        shutter = self.make_shutter()
        
        # Run code to be tested
        shutter.continue_gently_closing_if_applicable()

        # Check results
        self.assertFalse(shutter.thread._connection.ioloop.call_later.called)

        # Check if connection was not closed yet:
        # (because it is not applicable!)
        shutter.thread._connection.close.assert_not_called()

    def test_continuing_trigger(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter._ShutDowner__is_in_process_of_gently_closing = True
        shutter.recursive_decision_about_closing = mock.MagicMock() # to be able to check if it was called
        
        # Run code to be tested
        shutter.continue_gently_closing_if_applicable()
        self.assertTrue(shutter.thread._connection.ioloop.call_later.called)
        self.assertTrue(shutter.recursive_decision_about_closing.called)
        # Check if connection was closed:
        # (Cannot be closed, as we mock the method that would lead to closing)
        shutter.thread._connection.close.assert_not_called()

    def test_continuing_whole(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread.num_unpublished = 100
        shutter.thread.num_unconfirmed = 100
        shutter.thread.fake_confirms = False

        # Patch the shutter method to interrupt it:
        real_method = shutter.recursive_decision_about_closing
        shutter.recursive_decision_about_closing = mock.MagicMock() # to be able to check if it was called
        
        # Start a gentle close down, which is interrupted by the patch:
        shutter.finish_gently()

        # Shutting is interrupted, so messages were not sent:
        self.assertFalse(shutter.thread.add_event_publish_message.called)
        self.assertEqual(shutter.thread.num_unconfirmed, 100)
        self.assertEqual(shutter.thread.num_unpublished, 100)
        # The mock was called, instead of the real method (this is the interrupting):
        self.assertTrue(shutter.recursive_decision_about_closing.called)

        # Restore the old method
        shutter.recursive_decision_about_closing = real_method
        shutter.thread.fake_confirms = True

        # Run code to be tested: Continue shutting:
        shutter.continue_gently_closing_if_applicable()

        # Check if the shutting really was continued
        self.assertTrue(shutter.thread.add_event_publish_message.called)

        # Check if messages were sent now:
        self.assertTrue(shutter.thread.num_unconfirmed <= 99)
        self.assertTrue(shutter.thread.num_unpublished <= 99)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()

    #
    # Safety finish
    #


    def test_safety_finish_ok(self):

        # Preparation:
        shutter = self.make_shutter()
        shutter.thread.num_unconfirmed = 4
        shutter.thread.num_unpublished = 5

        # Run code to be tested:
        shutter.safety_finish()

        # Check result state:
        state = shutter.statemachine._StateMachine__state
        exp = shutter.statemachine._StateMachine__PERMANENTLY_UNAVAILABLE
        self.assertEqual(state, exp)

        # Check leftovers:
        self.assertEqual(shutter.thread.get_num_unconfirmed(), 4)
        self.assertEqual(shutter.thread.get_num_unpublished(), 5)
        # Check if connection was closed:
        shutter.thread._connection.close.assert_called()



import unittest
import logging
from esgfpid.assistant.consistency import Checker as Checker
import esgfpid.exceptions # need SolrError

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
import resources.TESTVALUES as TESTHELPERS

'''
Unit tests for esgfpid.assistant.consistency

This module needs a coupler. A coupler has references to
solr and RabbitMQ.

In these tests, we use a real coupler, but mock the solr
module with mock.Mock() objects that return the desired
things.
Access to RabbitMQ is never used, so the rabbit module is
not mocked.
 
'''
class ConsistencyTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def test_preparation_ok(self):

        # Preparations: Make test coupler with patched solr.
        prev = ['abc', 'def']
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev)

        # Args for the consistency checker:
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler

        # Run code to be tested:
        checker = Checker(**args)

        # Check result:
        self.assertIsInstance(checker, Checker,
            'Preparation failed.')
        self.assertTrue(checker.can_run_check(),
            'With previous files, check should be enabled, but isn\'t.')

    def test_check_ok(self):

        # Preparations: Make test coupler with patched solr.
        prev = ['abc', 'def']
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev)

        # Preparations: Make consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler
        checker = Checker(**args)

        # Run code to be tested:
        success = checker.data_consistency_check(prev)

        # Check result:
        self.assertTrue(success, 'The check should pass.')

    def test_check_too_many_files(self):

        # Preparations: Make test coupler with patched solr.
        prev = ['abc', 'def']
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev)

        # Preparations: Make consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler
        checker = Checker(**args)

        # Run code to be tested:
        too_many_files = prev+['ghi', 'jkl']
        success = checker.data_consistency_check(too_many_files)

        # Check result:
        self.assertFalse(success, 'The check should fail - too many files passed.')

    def test_check_too_few_files(self):

        # Preparations: Make test coupler with patched solr.
        prev = ['abc', 'def']
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev)

        # Preparations: Make consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler
        checker = Checker(**args)

        # Run code to be tested:
        too_few_files = ['abc']
        success = checker.data_consistency_check(too_few_files)

        # Check result:
        self.assertFalse(success, 'The check should fail - too few files passed.')

    def test_check_different_files(self):

        # Preparations: Make test coupler with patched solr.
        prev = ['abc', 'def']
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev)

        # Preparations: Make consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler
        checker = Checker(**args)

        # Run code to be tested:
        different_files = ['xyz', 'zyx']
        success = checker.data_consistency_check(different_files)

        # Check result:
        self.assertFalse(success, 'The check should fail - different files passed.')

    def test_check_no_files(self):

        # Preparations: Make test coupler with patched solr.
        prev = ['abc', 'def']
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev)

        # Preparations: Make consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler
        checker = Checker(**args)

        # Run code to be tested:
        no_files = []
        with self.assertRaises(ValueError):
            success = checker.data_consistency_check(no_files)

    def test_check_try_if_disabled(self):

        # Preparations: Make test coupler with patched solr.
        prev = ['abc', 'def']
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev)

        # Preparations: Make consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler
        checker = Checker(**args)
        # Switch it off!!!
        checker._Checker__will_run_check = False # Switch the check off!

        # Run code to be tested:
        with self.assertRaises(ValueError) as raised:
            success = checker.data_consistency_check(['any', 'thing'])
        self.assertIn('No reason specified', repr(raised.exception), 'Expected message did not arrive: "%s"' % repr(raised.exception))


    def test_preparation_files_is_empty_list(self):

        # Preparations: Make test coupler with patched solr.
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_empty_file_list(testcoupler)

        # Preparations: Make consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler
        checker = Checker(**args)

        # Run code to be tested:
        checker = Checker(**args)

        # Check result:
        self.assertIsInstance(checker, Checker,
            'Preparation failed.')
        self.assertFalse(checker.can_run_check(),
            'Without previous files, check should be disabled, but isn\'t.')

    def test_preparation_files_is_none(self):

        # Preparations: Make test coupler with patched solr.
        prev = None
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_previous_files(testcoupler, prev)

        # Preparations: Make consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler
        checker = Checker(**args)

        # Check result:
        self.assertIsInstance(checker, Checker,
            'Preparation failed.')
        self.assertFalse(checker.can_run_check(),
            'Without previous files, check should be disabled, but isn\'t.')

    def test_preparation_error(self):

        # Preparations: Make test coupler with patched solr.
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_raises_error(testcoupler, esgfpid.exceptions.SolrError)

        # Preparations: Args for consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler

        # Run code to be tested:
        checker = Checker(**args)

        # Check result:
        self.assertIsInstance(checker, Checker,
            'Preparation failed.')
        self.assertFalse(checker.can_run_check(),
            'Without previous files, check should be disabled, but isn\'t.')

    def test_preparation_switched_off(self):

        # Preparations: Make test coupler with switched off solr.
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)

        # Preparations: Args for consistency checker.
        args = TESTHELPERS.get_args_for_consistency_check()
        args['coupler'] = testcoupler

        # Run code to be tested:
        checker = Checker(**args)

        # Check result:
        self.assertIsInstance(checker, Checker,
            'Preparation failed.')
        self.assertFalse(checker.can_run_check(),
            'Without previous files, check should be disabled, but isn\'t.')


# Expected error messages for the unit tests
# of the check.py module.

expected_message_ok1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication and connection (this.is.my.only.host)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... ok.
 .. checking exchange ...
 .. checking exchange ... ok.
Config for PID module (rabbit messaging queue).. ok.
Successful connection to PID messaging queue at "this.is.my.only.host".'''

expected_message_ok2 = None # no message returned when success occurs!

expected_message_plus_testmsg_ok1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication and connection (this.is.my.only.host)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... ok.
 .. checking exchange ...
 .. checking exchange ... ok.
 .. checking message ...
 .. checking message ... ok.
Config for PID module (rabbit messaging queue).. ok.
Successful connection to PID messaging queue at "this.is.my.only.host".'''

expected_message_plus_testmsg_ok2 = None # no message returned when success occurs!

expected_message_first_connection_failed_then_ok1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking connection (this.is.my.favourite.host)... FAILED.
 .. giving this node a lower priority..
 .. checking connection (mystery-tour.uk)... FAILED.
 .. giving this node a lower priority..
 .. checking authentication and connection (tomato.salad-with-spam.fr)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... ok.
 .. checking exchange ...
 .. checking exchange ... ok.
Config for PID module (rabbit messaging queue).. ok.
Successful connection to PID messaging queue at "tomato.salad-with-spam.fr".'''

expected_message_first_connection_failed_then_ok2 = None # no message returned when success occurs!

expected_message_connection_failed_unknown1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking connection (this.is.my.only.host)... FAILED.
 .. giving this node a lower priority..
Config for PID module (rabbit messaging queue) .. FAILED!'''

expected_message_connection_failed_unknown2 = '''********************************************************************
*** PROBLEM IN SETTING UP                                        ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                          ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:   ***
***  - host "this.is.my.only.host": Unknown connection failure.  ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE. ***
********************************************************************'''


expected_message_connection_failed_only1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking connection (this.is.my.only.host)... FAILED.
 .. giving this node a lower priority..
Config for PID module (rabbit messaging queue) .. FAILED!'''

expected_message_connection_failed_only2 = '''*********************************************************************************
*** PROBLEM IN SETTING UP                                                     ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                                       ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:                ***
***  - host "this.is.my.only.host": Connection failure (wrong host or port?). ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE.              ***
*********************************************************************************'''

expected_message_connection_failed_several1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking connection (this.is.my.favourite.host)... FAILED.
 .. giving this node a lower priority..
 .. checking connection (mystery-tour.uk)... FAILED.
 .. giving this node a lower priority..
 .. checking connection (tomato.salad-with-spam.fr)... FAILED.
 .. giving this node a lower priority..
Config for PID module (rabbit messaging queue) .. FAILED!'''

expected_message_connection_failed_several2 = '''**************************************************************************************
*** PROBLEM IN SETTING UP                                                          ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                                            ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:                     ***
***  - host "this.is.my.favourite.host": Connection failure (wrong host or port?). ***
***  - host "mystery-tour.uk": Connection failure (wrong host or port?).           ***
***  - host "tomato.salad-with-spam.fr": Connection failure (wrong host or port?). ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE.                   ***
**************************************************************************************'''

expected_message_authentication_failed_only1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication (this.is.my.only.host)... FAILED.
 .. giving this node a lower priority..
Config for PID module (rabbit messaging queue) .. FAILED!'''

expected_message_authentication_failed_only2 = '''*************************************************************************************************
*** PROBLEM IN SETTING UP                                                                     ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                                                       ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:                                ***
***  - host "this.is.my.only.host": Authentication failure (user johndoe, password abc123yx). ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE.                              ***
*************************************************************************************************'''

expected_message_authentication_failed_several1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication (this.is.my.favourite.host)... FAILED.
 .. giving this node a lower priority..
 .. checking authentication (mystery-tour.uk)... FAILED.
 .. giving this node a lower priority..
 .. checking authentication (tomato.salad-with-spam.fr)... FAILED.
 .. giving this node a lower priority..
Config for PID module (rabbit messaging queue) .. FAILED!'''

expected_message_authentication_failed_several2 = '''******************************************************************************************************
*** PROBLEM IN SETTING UP                                                                          ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                                                            ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:                                     ***
***  - host "this.is.my.favourite.host": Authentication failure (user johndoe, password abc123yx). ***
***  - host "mystery-tour.uk": Authentication failure (user johndoe, password abc123yx).           ***
***  - host "tomato.salad-with-spam.fr": Authentication failure (user johndoe, password abc123yx). ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE.                                   ***
******************************************************************************************************'''


expected_message_channel_closed_only1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication and connection (this.is.my.only.host)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... FAILED.
 .. giving this node a lower priority..
Config for PID module (rabbit messaging queue) .. FAILED!'''

expected_message_channel_closed_only2 = '''********************************************************************
*** PROBLEM IN SETTING UP                                        ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                          ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:   ***
***  - host "this.is.my.only.host": Channel failure.             ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE. ***
********************************************************************'''


expected_message_first_channel_closed_then_ok1 = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication and connection (this.is.my.favourite.host)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... FAILED.
 .. giving this node a lower priority..
 .. checking authentication and connection ...
 .. checking authentication and connection (mystery-tour.uk)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... FAILED.
 .. giving this node a lower priority..
 .. checking authentication and connection ...
 .. checking authentication and connection (tomato.salad-with-spam.fr)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... ok.
 .. checking exchange ...
 .. checking exchange ... ok.
Config for PID module (rabbit messaging queue).. ok.
Successful connection to PID messaging queue at "tomato.salad-with-spam.fr".'''

expected_message_first_channel_closed_then_ok2 = None



import os
import os.path

def get_boolean(arg):
    if isinstance(arg, bool):
        return arg
    elif isinstance(arg, str):
        if arg.lower() in ['true']:
            return True
        elif arg.lower() in ['false']:
            return False
        else:
            raise ValueError('%s (string) cannot be parsed to a boolean.', arg)
    else:
        raise ValueError('%s (%s) cannot be parsed to a boolean.', arg, type(arg))

def ensure_directory_exists(path):
    try: 
        os.mkdir(path)
    except OSError as e:
        if 'File exists' in repr(e):
            if not os.path.isdir(path):
                raise




import logging
import copy
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn, log_every_x_times
from .exceptions import UnknownServerResponse

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''
=========
Confirmer
=========

The confirmer is responsible for handling the confirms sent by RabbitMQ.
This is a quite simple module, it only reacts and does not trigger any
other actions itself.

It has a stack of unconfirmed messages (which is filled by the feeder,
it puts every message it has successfully published into that stack).

For each confirm, it must check what kind of confirm it is (ack/nack, single/multiple),
and act accordingly. Confirmed messages  and their delivery numbers must be deleted
from the stack. Unconfirmed messages remain. Nacked messages are stored in an
extra stack.

The unconfirmed messages can be retrieved from the confirmer to be republished.

API:
 * on_delivery_confirmation() is called by RabbitMQ.
 * reset_unconfirmed_messages_and_delivery_tags() called by builder, during reconnection
 * get_unconfirmed_messages_as_list_copy() called by builder, during reconnection
 * put_to_unconfirmed_delivery_tags() called by feeder, to fill the stack
 * put_to_unconfirmed_messages_dict() called by feeder, to fill the stack

'''

class Confirmer(object):

    def __init__(self):

        # Logging:
        self.__first_confirm_receival = True
        self.__logcounter = 1
        self.__LOGFREQUENCY = 10

        # Stacks of unconfirmed/nacked messages:
        self.__unconfirmed_delivery_tags = [] # only accessed internally
        self.__unconfirmed_messages_dict = {} # dict, because I need to retrieve them by delivery tag (on ack/nack)
        self.__nacked_messages = []           # only accessed internally, and from outside after thread is dead

    '''
    Callback, called by RabbitMQ.
    '''
    def on_delivery_confirmation(self, method_frame):
        deliv_tag, confirmation_type, multiple = self.__get_confirm_info(method_frame)
        log_every_x_times(LOGGER, self.__logcounter, self.__LOGFREQUENCY, 'Received a confirm (%s)', confirmation_type)
        self.__logcounter += 1

        if confirmation_type == 'ack':
            logtrace(LOGGER, 'Received "ACK" from messaging service.')
            self.__react_on_ack(deliv_tag, multiple)
        elif confirmation_type == 'nack':
            logtrace(LOGGER, 'Received "NACK" from messaging service.')
            self.__react_on_nack(deliv_tag, multiple)
        else:
            msg = 'Received asynchronous response of unknown type from messaging service.'
            logwarn(LOGGER, msg)
            raise UnknownServerResponse(msg+':'+str(method_frame))
            # This should never happen, unless if I parse the server's response wrongly.

    def __react_on_ack(self, deliv_tag, multiple):
        if self.__first_confirm_receival:
            self.__first_confirm_receival = False
            loginfo(LOGGER, 'Received first message confirmation from RabbitMQ.')

        if multiple:
            logtrace(LOGGER, 'Received "ACK" for multiple messages from messaging service.')
            self.__react_on_multiple_delivery_ack(deliv_tag)
        else:
            logtrace(LOGGER, 'Received "ACK" for single message from messaging service.')
            self.__react_on_single_delivery_ack(deliv_tag)


    def __react_on_nack(self, deliv_tag, multiple):
        if multiple:
            logwarn(LOGGER, 'Received "NACK" for delivery tag: %i and below.', deliv_tag)
            self.__nack_delivery_tag_and_message_several(deliv_tag)
        else:
            logwarn(LOGGER, 'Received "NACK" for delivery tag: %i.', deliv_tag)
            self.__nack_delivery_tag_and_message_single(deliv_tag)

    def __nack_delivery_tag_and_message_single(self, deliv_tag):
        msg = self.__unconfirmed_messages_dict.pop(str(deliv_tag))
        self.__nacked_messages.append(msg)
        self.__unconfirmed_delivery_tags.remove(deliv_tag)

    def __nack_delivery_tag_and_message_several(self, deliv_tag):
        for candidate_deliv_tag in copy.copy(self.__unconfirmed_delivery_tags):
            if candidate_deliv_tag <= deliv_tag:
                self.__nack_delivery_tag_and_message_single(candidate_deliv_tag)

    def __get_confirm_info(self, method_frame):
        try:
            deliv_tag = method_frame.method.delivery_tag # integer
            confirmation_type = method_frame.method.NAME.split('.')[1].lower() # "ack" or "nack" or ...
            multiple = method_frame.method.multiple # Boolean
            return deliv_tag, confirmation_type, multiple
        except AttributeError as e:
            raise UnknownServerResponse(str(method_frame)+' - '+repr(e))
        except IndexError as e:
            raise UnknownServerResponse(str(method_frame)+' - '+repr(e))

    def __react_on_single_delivery_ack(self, deliv_tag):
        self.__remove_delivery_tag_and_message_single(deliv_tag)
        logdebug(LOGGER, 'Received ack for delivery tag %i. Waiting for %i confirms.', deliv_tag, len(self.__unconfirmed_delivery_tags))
        logtrace(LOGGER, 'Received ack for delivery tag %i.', deliv_tag)
        logtrace(LOGGER, 'Now left in queue to be confirmed: %i messages.', len(self.__unconfirmed_delivery_tags))

    def __react_on_multiple_delivery_ack(self, deliv_tag):
        self.__remove_delivery_tag_and_message_several(deliv_tag)
        logdebug(LOGGER, 'Received ack for delivery tag %i and all below. Waiting for %i confirms.', deliv_tag, len(self.__unconfirmed_delivery_tags))
        logtrace(LOGGER, 'Received ack for delivery tag %i and all below.', deliv_tag)
        logtrace(LOGGER, 'Now left in queue to be confirmed: %i messages.', len(self.__unconfirmed_delivery_tags))

    def __remove_delivery_tag_and_message_single(self, deliv_tag):
        try:
            self.__unconfirmed_delivery_tags.remove(deliv_tag)
            ms = self.__unconfirmed_messages_dict.pop(str(deliv_tag))
            logtrace(LOGGER, 'Received ack for message %s.', ms)
        except ValueError as e:
            logdebug(LOGGER, 'Could not remove %i from unconfirmed.', deliv_tag)

    def __remove_delivery_tag_and_message_several(self, deliv_tag):
        for candidate_deliv_tag in copy.copy(self.__unconfirmed_delivery_tags):
            if candidate_deliv_tag <= deliv_tag:
                self.__remove_delivery_tag_and_message_single(candidate_deliv_tag)


    ''' Called by unit test.'''
    def get_num_unconfirmed(self):
        return len(self.__unconfirmed_messages_dict)

    ''' Called by unit test.'''
    def get_copy_of_unconfirmed_tags(self):
        return self.__unconfirmed_delivery_tags[:]

    '''
    Called by the main thread, for rescuing, after joining.
    And by unit test.
    '''
    def get_copy_of_nacked(self):
        return self.__nacked_messages[:]

    '''
    Called by feeder, to let the confirmer know which had been sent.
    '''
    def put_to_unconfirmed_delivery_tags(self, delivery_tag):
        logtrace(LOGGER, 'Adding delivery tag %i to unconfirmed.', delivery_tag)
        self.__unconfirmed_delivery_tags.append(delivery_tag)

    '''
    Called by feeder, to let the confirmer know which had been sent.
    '''
    def put_to_unconfirmed_messages_dict(self, delivery_tag, msg):
        logtrace(LOGGER, 'Adding message with delivery tag %i to unconfirmed: %s', delivery_tag, msg)
        self.__unconfirmed_messages_dict[str(delivery_tag)] = msg

    '''
    This resets which messages had not be confirmed yet.
    
    IMPORTANT:
    Before doing this, retrieve the unconfirmed messages
    and republish them! Otherwise, they will be lost!
    (We do not do this here, as we have no reference to
    the builder module, to avoid circular references).

    Called by builder, during reconnection.
    After a reconnection, no more confirms can be received,
    so the messages that are unconfirmed now cannot be
    confirmed any more.

    Any new confirms that arrive after a reconnection will
    carry new delivery tags (as they start over at one),
    so we need to reset the old ones, otherwise we'll take
    the wrong messages as confirmed.

    From RabbitMQ docs:
    "The server-assigned and channel-specific delivery tag. The delivery tag is valid only within the channel from which the message was received. [...] The server MUST NOT use a zero value for delivery tags."
    See: https://www.rabbitmq.com/amqp-0-9-1-reference.html

    '''
    def reset_unconfirmed_messages_and_delivery_tags(self):
        self.__unconfirmed_delivery_tags = []
        self.__unconfirmed_messages_dict = {}

    '''
    Called by builder, during reconnection,
    to rescue these unconfirmed messages, in order to
    republish them, as soon as a new connection is
    available.

    Also called by the main thread, for rescuing messages
    after joining.

    As dict objects are not thread-safe, better call
    this only after joining.
    '''
    def get_unconfirmed_messages_as_list_copy(self):
        newlist = []
        for deliv_tag,message in self.__unconfirmed_messages_dict.items():
            newlist.append(message)
        return newlist



import logging
import requests
import json
import esgfpid.utils
import esgfpid.solr.tasks.filehandles_same_dataset
import esgfpid.solr.tasks.all_versions_of_dataset
import esgfpid.solr.serverconnector
import esgfpid.defaults
import esgfpid.exceptions
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''
This class provides the solr facade for the rest of
the library.

All requests to solr are addressed to an instance of
this class.

It redirects the calls for the various tasks to the
responsible submodules, which create solr queries and
can parse results (:py:mod:`~esgfpid.solr.tasks`)

The actual interaction with solr server (sending of
a ready-made query, receiving the response) is handled 
by another class
(:py:class:`~esgfpid.solr.serverconnector.SolrServerConnector`).

'''
class SolrInteractor(object):

    # Constructor:

    '''
    :param switched_off: Mandatory. Boolean.
    :param prefix: Mandatory if not switched off.
    :param solr_url: Mandatory if not switched off.
    :param https_verify: Mandatory if not switched off.
    :param disable_insecure_request_warning: Mandatory if not switched off.
    '''
    def __init__(self, **args):

        mandatory_args = [
            'switched_off',
            'prefix',
            'solr_url',
            'https_verify',
            'disable_insecure_request_warning'
        ]
        esgfpid.utils.check_presence_of_mandatory_args(args, ['switched_off'])

        if args['switched_off'] == True:
            logdebug(LOGGER, 'Initializing solr module without access..')
            self.__init_without_access()
            logdebug(LOGGER, 'Initializing solr module without access.. done')

        else:
            esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
            logdebug(LOGGER, 'Initializing solr module..')
            self.__init_with_access(args)
            logdebug(LOGGER, 'Initializing solr module.. done')

    def __init_without_access(self):
        self.__switched_on = False
        self.__prefix = None
        self.__solr_server_connector = None

    def __init_with_access(self, args):
        self.__switched_on = True
        self.__check_presence_of_args(args)
        self.__prefix = args['prefix']
        self.__make_server_connector(args)

    def __check_presence_of_args(self, args):
        mandatory_args = ['solr_url', 'prefix', 'https_verify',
            'disable_insecure_request_warning', 'switched_off']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

    def __make_server_connector(self, args):
        self.__solr_server_connector = esgfpid.solr.serverconnector.SolrServerConnector(
            solr_url = args['solr_url'],
            https_verify = args ['https_verify'],
            disable_insecure_request_warning = args['disable_insecure_request_warning']
        )

    # Getter

    '''
    State getter.

    :returns: True if the solr module is switched off, i.e.
        it either received a switch-off flag from the library
        or had no solr URL passed. False if not switched off.
    '''
    def is_switched_off(self):
        return not self.__switched_on

    # Methods called by tasks:

    def send_query(self, query):
        ''' This method is called by the tasks. It is redirected to the submodule.'''
        if self.__switched_on:
            return self.__solr_server_connector.send_query(query)
        else:
            msg = 'Not sending query'
            LOGGER.debug(msg)
            raise esgfpid.exceptions.SolrSwitchedOff(msg)

    def make_solr_base_query(self):
        query_dict = {}
        query_dict['distrib'] = esgfpid.defaults.SOLR_QUERY_DISTRIB
        query_dict['format'] = 'application/solr+json'
        query_dict['limit'] = 0 # As we don't want all the details of the found files/datasets!
        return query_dict

    #####################
    ### Various tasks ###
    #####################

    # Task 1

    def retrieve_file_handles_of_same_dataset(self, **args):
        '''
        :return: List of handles, or empty list. Should never return None.
        :raise: SolrSwitchedOff
        :raise SolrError: If both strategies to find file handles failed.
        '''

        mandatory_args = ['drs_id', 'version_number', 'data_node']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        LOGGER.debug('Looking for files of dataset "%s", version "%s".',
                     args['drs_id'], str(args['version_number']))

        if self.__switched_on:
            return self.__retrieve_file_handles_of_same_dataset(**args)
        else:
            msg = 'Cannot retrieve handles of files of the same dataset.'
            raise esgfpid.exceptions.SolrSwitchedOff(msg)

    def __retrieve_file_handles_of_same_dataset(self, **args):
        finder = esgfpid.solr.tasks.filehandles_same_dataset.FindFilesOfSameDatasetVersion(self)
        args['prefix'] = self.__prefix
        file_handles = finder.retrieve_file_handles_of_same_dataset(**args)
        return file_handles

    # Task 2

    def retrieve_datasethandles_or_versionnumbers_of_allversions(self, drs_id):
        LOGGER.debug('Looking for dataset handles or version numbers of '+
                     'dataset "%s".', drs_id)

        if self.__switched_on:
            return self.__retrieve_datasethandles_or_versionnumbers_of_allversions(drs_id)
        else:
            msg = 'Cannot retrieve handles or version numbers of all versions of the dataset.'
            raise esgfpid.exceptions.SolrSwitchedOff(msg)

    def __retrieve_datasethandles_or_versionnumbers_of_allversions(self, drs_id):
        finder = esgfpid.solr.tasks.all_versions_of_dataset.FindVersionsOfSameDataset(self)
        result_dict = finder.retrieve_dataset_handles_or_version_numbers_of_all_versions(drs_id, self.__prefix)
        return result_dict









import esgfpid.rabbit.rabbit
import logging
import time
import datetime

TESTVALUES_REAL = dict(
    url_messaging_service='handle-esgf.dkrz.de',
    messaging_exchange='rabbitsender_integration_tests',
    exchange_no_queue='rabbitsender_integration_tests_no_queue',
    exchange_name='rabbitsender_integration_tests', # TODO JUST ONE NAME!
    rabbit_username='sendertests',
    rabbit_password='testmebaby',
    routing_key='test_routing_key',
)

#logging.basicConfig(level=logging.DEBUG)
logdate = datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
FORMAT = "%(asctime)-15s %(levelname)s:%(name)s:%(message)s"  
logging.basicConfig(level=logging.DEBUG, filename='test_log_'+logdate+'.log', filemode='w', format=FORMAT)
logging.getLogger('esgfpid.rabbit.asynchronous.thread_builder').setLevel(logging.DEBUG)
logging.getLogger('pika').setLevel(logging.DEBUG)

# Test variables:
routing_key = TESTVALUES_REAL['routing_key']
num_messages = 150

# Make connection:
#testrabbit = self.make_rabbit_with_access()
testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(
      exchange_name=123,
      urls_fallback=TESTVALUES_REAL['url_messaging_service'],
      url_preferred=None,
      username=TESTVALUES_REAL['rabbit_username'],
      password=TESTVALUES_REAL['rabbit_password'],
)
testrabbit.start()

# Run code to be tested:
for i in range(num_messages):
      print("Publishing message %i..." % i)
      testrabbit.send_message_to_queue({"stuffee":"foo"+str(i+1),"ROUTING_KEY":routing_key})
      print("Publishing message %i... done." % i)
      time.sleep(3)


# Close connection after some time:
#print('Wait to close... 1')
#time.sleep(5)
print('Try to close... 1')
testrabbit.finish()

# Check result:
leftovers = testrabbit.get_leftovers()
if not len(leftovers)==0:
      print('There is leftovers: '+str(len(leftovers)))
if not testrabbit.is_finished():
      print('Thread is still alive.')

# Close connection after some time:
print('Wait to close... 2')
time.sleep(5)
print('Try to close... 2')
testrabbit.finish()

# Check result:
leftovers = testrabbit.get_leftovers()
if not len(leftovers)==0:
      print('There is leftovers: '+str(len(leftovers)))
if not testrabbit.is_finished():
      print('Thread is still alive.')


# Close connection after some time:
print('Wait to close... 2')
time.sleep(5)
print('Try to force-close... 2')
testrabbit.force_finish()
print('Force-close done')

# Check result:
leftovers = testrabbit.get_leftovers()
if not len(leftovers)==0:
      print('There is leftovers: '+str(len(leftovers)))
if not testrabbit.is_finished():
      print('Thread is still alive.')




import unittest
import mock
import logging
import json
import esgfpid.solr.tasks.utils
import esgfpid.exceptions
import tests.utils

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Load some data that is needed for testing
PATH_RES = tests.utils.get_super_neighbour_directory(__file__, 'resources')
print(PATH_RES)
SOLR_RESPONSE = json.load(open(PATH_RES+'/solr_response.json'))

class SolrUtilsTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def test_extract_handles_from_response_json_ok(self):
        
        # Test variables
        response_json = SOLR_RESPONSE
        prefix = '10876.test'
  
        # Run code to be tested:
        received_handles = esgfpid.solr.tasks.utils.extract_file_handles_from_response_json(response_json, prefix)

        # Check result:
        expected_handles = [
            'hdl:10876.test/fe8b7a91-1391-4e36-a594-77d63f367846',
            'hdl:10876.test/045c78f2-90f5-465d-af68-54a23d7607d4',
            'hdl:10876.test/080c4b28-98ff-40ac-8d8f-76bea0db014e',
            'hdl:10876.test/079460fe-2ab8-461b-8bec-15b2028290a1',
            'hdl:10876.test/08bc91b7-6397-460c-977d-e3b4d3d185c6',
            'hdl:10876.test/04ebf501-0d5c-4207-b261-2507331fe081',
            'hdl:10876.test/e4f9bac0-f8ff-453f-a634-9953932801f6', 
            'hdl:10876.test/05935211-c8cc-4bf2-8149-745f41a7d07c'
        ]
        equal = (sorted(received_handles) == sorted(expected_handles))
        self.assertTrue(equal, 'Expected: %s\nReceived: %s' % (expected_handles, received_handles))

    def test_no_handles_found(self):

        # Test variables
        prefix = 'whatever'
        response_json = {
            "bla": "blub",
            "facet_counts": {
                "facet_queries": { },
                "facet_fields": {
                    "handle": [ ],
                    "pid": [ ],
                    "tracking_id": [ ]
                },
            "facet_dates": { },
            "facet_ranges": { },
            "facet_intervals": { },
            "facet_heatmaps": { }
            }
        }

        # Run code to be tested:
        received_handles = esgfpid.solr.tasks.utils.extract_file_handles_from_response_json(response_json, prefix)

        # Check result:
        expected_handles = [] # None were found!
        equal = (received_handles == expected_handles)
        self.assertTrue(equal, 'Expected: %s\nReceived: %s' % (expected_handles, received_handles))

    def test_handles_found_ok(self):

        # Test variables
        prefix = 'myprefix'
        response_json = {
            "bla": "blub",
            "facet_counts": {
                "facet_queries": { },
                "facet_fields": {
                    "handle": [ ],
                    "pid": [ ],
                    "tracking_id": [
                        "hdl:myprefix/387b00d1-fe4e-40e3-a654-2dffbcf5bec6",
                        1,
                        "hdl:myprefix/79352fe9-1fbe-48b7-9b44-d6b36312296a",
                        1,
                        "hdl:myprefix/b9667250-b15c-47cf-b79c-5ab9fd49e5e1",
                        1
                    ]
                },
            "facet_dates": { },
            "facet_ranges": { },
            "facet_intervals": { },
            "facet_heatmaps": { }
            }
        }

        # Run code to be tested:
        received_handles = esgfpid.solr.tasks.utils.extract_file_handles_from_response_json(response_json, prefix)

        # Check result:
        expected_handles = ["hdl:myprefix/387b00d1-fe4e-40e3-a654-2dffbcf5bec6",
                            "hdl:myprefix/79352fe9-1fbe-48b7-9b44-d6b36312296a",
                            "hdl:myprefix/b9667250-b15c-47cf-b79c-5ab9fd49e5e1"]
        equal = (set(received_handles) == set(expected_handles))
        self.assertEqual(len(received_handles), len(expected_handles), 'Not the same number of handles')
        self.assertTrue(equal, 'Expected: %s\nReceived: %s' % (expected_handles, received_handles))

    def test_handles_found_without_prefix(self):
        ''' Prefix (and hdl:) should be put in front of handle! '''

        # Test variables
        prefix = 'myprefix'
        response_json = {
            "bla": "blub",
            "facet_counts": {
                "facet_queries": { },
                "facet_fields": {
                    "handle": [ ],
                    "pid": [ ],
                    "tracking_id": [
                        "387b00d1-fe4e-40e3-a654-2dffbcf5bec6",
                        1,
                        "79352fe9-1fbe-48b7-9b44-d6b36312296a",
                        1,
                        "b9667250-b15c-47cf-b79c-5ab9fd49e5e1",
                        1
                    ]
                },
            "facet_dates": { },
            "facet_ranges": { },
            "facet_intervals": { },
            "facet_heatmaps": { }
            }
        }

        # Run code to be tested:
        received_handles = esgfpid.solr.tasks.utils.extract_file_handles_from_response_json(response_json, prefix)

        # Check result:
        expected_handles = ["hdl:myprefix/387b00d1-fe4e-40e3-a654-2dffbcf5bec6",
                            "hdl:myprefix/79352fe9-1fbe-48b7-9b44-d6b36312296a",
                            "hdl:myprefix/b9667250-b15c-47cf-b79c-5ab9fd49e5e1"]
        equal = (set(received_handles) == set(expected_handles))
        self.assertEqual(len(received_handles), len(expected_handles), 'Not the same number of handles')
        self.assertTrue(equal, 'Expected: %s\nReceived: %s' % (expected_handles, received_handles))

    def test_handles_found_without_hdl(self):
        ''' hdl: should be put in front of the handle names!'''

        # Test variables
        prefix = 'myprefix'
        response_json = {
            "bla": "blub",
            "facet_counts": {
                "facet_queries": { },
                "facet_fields": {
                    "handle": [ ],
                    "pid": [ ],
                    "tracking_id": [
                        "myprefix/387b00d1-fe4e-40e3-a654-2dffbcf5bec6",
                        1,
                        "myprefix/79352fe9-1fbe-48b7-9b44-d6b36312296a",
                        1,
                        "myprefix/b9667250-b15c-47cf-b79c-5ab9fd49e5e1",
                        1
                    ]
                },
            "facet_dates": { },
            "facet_ranges": { },
            "facet_intervals": { },
            "facet_heatmaps": { }
            }
        }

        # Run code to be tested:
        received_handles = esgfpid.solr.tasks.utils.extract_file_handles_from_response_json(response_json, prefix)

        # Check result:
        expected_handles = ["hdl:myprefix/387b00d1-fe4e-40e3-a654-2dffbcf5bec6",
                            "hdl:myprefix/79352fe9-1fbe-48b7-9b44-d6b36312296a",
                            "hdl:myprefix/b9667250-b15c-47cf-b79c-5ab9fd49e5e1"]
        equal = (set(received_handles) == set(expected_handles))
        self.assertEqual(len(received_handles), len(expected_handles), 'Not the same number of handles')
        self.assertTrue(equal, 'Expected: %s\nReceived: %s' % (expected_handles, received_handles))

    def test_no_field_tracking_id(self):

        # Test variables
        prefix = 'myprefix'
        response_json = {
            "bla": "blub",
            "facet_counts": {
                "facet_queries": { },
                "facet_fields": {
                    "handle": [ ],
                    "pid": [ ]
                },
            "facet_dates": { },
            "facet_ranges": { },
            "facet_intervals": { },
            "facet_heatmaps": { }
            }
        }

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.SolrResponseError):
            received_handles = esgfpid.solr.tasks.utils.extract_file_handles_from_response_json(response_json, prefix)


    def test_no_facet_fields(self):

        # Test variables
        prefix = 'myprefix'
        response_json = {
            "bla": "blub",
            "facet_counts": {
                "facet_queries": { }
                },
            "facet_dates": { },
            "facet_ranges": { },
            "facet_intervals": { },
            "facet_heatmaps": { }
        }

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.SolrResponseError):
            received_handles = esgfpid.solr.tasks.utils.extract_file_handles_from_response_json(response_json, prefix)

    def test_response_is_none(self):

        # Test variables
        prefix = 'myprefix'
        response_json = None

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.SolrResponseError):
            received_handles = esgfpid.solr.tasks.utils.extract_file_handles_from_response_json(response_json, prefix)



import unittest
import mock
import logging
import json
import tests.resources.pikamock
import tests.main_test_script
import pika.exceptions
import pika
import tests.globalvar

import esgfpid.rabbit
from esgfpid.rabbit.exceptions import PIDServerException

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

# Skip slow tests?
import globalvar
if globalvar.QUICK_ONLY:
    print('Skipping slow tests in module "%s".' % __name__)


class RabbitConnectorTestCase(unittest.TestCase):

    slow_message = '\nRunning a slow test (avoid by using -ls flag).'


    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)
    
    def tearDown(self):
        LOGGER.info('#############################')

    #
    # Setup of the connection/channel:
    #

    def test_init_ok(self):

        # Test variables:
        nodemanager = TESTHELPERS.get_nodemanager()

        # Run code to be tested:
        testrabbit = esgfpid.rabbit.synchronous.SynchronousRabbitConnector(nodemanager)

        # Check results:
        # Instance is created:
        self.assertIsInstance(testrabbit, esgfpid.rabbit.synchronous.SynchronousRabbitConnector)
        # Communcation not established yet:
        established = testrabbit._SynchronousRabbitConnector__communication_established
        self.assertFalse(established)

    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_open_rabbit_connection_ok(self, connectionmock):

        # Define replacement for mocked method:
        # Using side_effect instead of return_value, so we can use the params object passed as arg!
        def make_mocked_connection(params):
            return tests.resources.pikamock.MockPikaBlockingConnection(params)
        connectionmock.side_effect = make_mocked_connection

        # Make test rabbit:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()
        
        # Run code to be tested (during this, the connection is opened, i.e. the patch is needed)
        testrabbit.open_rabbit_connection()

        # Check result:
        # Connection created:
        conn = testrabbit._SynchronousRabbitConnector__connection
        self.assertIsInstance(conn, tests.resources.pikamock.MockPikaBlockingConnection)
        # Channel created:
        channel = testrabbit._SynchronousRabbitConnector__channel
        self.assertIsInstance(channel, tests.resources.pikamock.MockChannel)
        # Communcation established:
        established = testrabbit._SynchronousRabbitConnector__communication_established
        self.assertTrue(established)
        # Connected to preferred URL:
        expected_host = RABBIT_URL_TRUSTED
        self.assertIn(conn.host, expected_host, 'Connected to: %s instead of %s' % (conn.host, expected_host))


    @unittest.skipIf(globalvar.QUICK_ONLY, '(this test is slow)')
    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_open_rabbit_connection_not_ok(self, connectionmock):

        print(self.slow_message)

        # Define replacement for mocked method:
        # All connection attempts fail with an Authentication Error.
        connectionmock.side_effect = pika.exceptions.AuthenticationError

        # Make test rabbit:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()
        
        # Run code to be tested (during this, the connection is opened, i.e. the patch is needed)
        with self.assertRaises(PIDServerException) as e:
            testrabbit.open_rabbit_connection()

        # Check result:
        # Connection is None:
        conn = testrabbit._SynchronousRabbitConnector__connection
        self.assertIsNone(conn)
        # Channel is None:
        channel = testrabbit._SynchronousRabbitConnector__channel
        self.assertIsNone(channel)
        # Communcation not established:
        established = testrabbit._SynchronousRabbitConnector__communication_established
        self.assertFalse(established)


    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_open_rabbit_connection_first_fails_auth(self, connectionmock):

        # Define replacement for mocked object:
        # RabbitMQ with priority 1 fails, the one with
        # priority 2 succeeds.
        def make_mocked_connection(params):
            if params.host == RABBIT_URL_TRUSTED:
                raise pika.exceptions.ProbableAuthenticationError
            elif params.host == RABBIT_URL_TRUSTED2:
                return tests.resources.pikamock.MockPikaBlockingConnection(params)
            else:
                raise ValueError("Unexpected RabbitMQ URL. Unit tests cannot work properly if the wrong test data is passed.")
        connectionmock.side_effect = make_mocked_connection

        # Make test rabbit:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()

        # Run code to be tested (during this, the connection is opened, i.e. the patch is needed)
        testrabbit.open_rabbit_connection()

        # Check result:
        # Connection created:
        conn = testrabbit._SynchronousRabbitConnector__connection
        self.assertIsInstance(conn, tests.resources.pikamock.MockPikaBlockingConnection)
        # Channel created:
        channel = testrabbit._SynchronousRabbitConnector__channel
        self.assertIsInstance(channel, tests.resources.pikamock.MockChannel)
        # Communcation established:
        established = testrabbit._SynchronousRabbitConnector__communication_established
        self.assertTrue(established)
        # Connected to fallback URL:
        expected_host = RABBIT_URL_TRUSTED2
        self.assertIn(conn.host, expected_host, 'Connected to: %s instead of %s' % (conn.host, expected_host))


    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_open_rabbit_connection_first_fails_conn(self, connectionmock):

        # Define replacement for mocked method:
        # RabbitMQ with priority 1 fails, the one with
        # priority 2 succeeds.
        def make_mocked_connection(params):
            if params.host == RABBIT_URL_TRUSTED:
                raise pika.exceptions.AMQPConnectionError
            elif params.host == RABBIT_URL_TRUSTED2:
                return tests.resources.pikamock.MockPikaBlockingConnection(params)
            else:
                raise ValueError("Unexpected RabbitMQ URL. Unit tests cannot work properly if the wrong test data is passed.")
        connectionmock.side_effect = make_mocked_connection

        # Make test rabbit:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()

        # Run code to be tested (during this, the connection is opened, i.e. the patch is needed)
        testrabbit.open_rabbit_connection()

        # Check result:
        # Connection created:
        conn = testrabbit._SynchronousRabbitConnector__connection
        self.assertIsInstance(conn, tests.resources.pikamock.MockPikaBlockingConnection)
        # Channel created:
        channel = testrabbit._SynchronousRabbitConnector__channel
        self.assertIsInstance(channel, tests.resources.pikamock.MockChannel)
        # Communcation established:
        established = testrabbit._SynchronousRabbitConnector__communication_established
        self.assertTrue(established)
        # Connected to fallback URL:
        expected_host = RABBIT_URL_TRUSTED2
        self.assertIn(conn.host, expected_host, 'Connected to: %s instead of %s' % (conn.host, expected_host))

    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_open_rabbit_connection_first_two_fail(self, connectionmock):

        # Define replacement for mocked method:
        # RabbitMQ with priority 1&2 fail, the third one succeeds.
        def make_mocked_connection(params):
            if params.host == RABBIT_URL_TRUSTED or params.host == RABBIT_URL_TRUSTED2:
                raise pika.exceptions.AMQPConnectionError
            elif params.host == RABBIT_URL_TRUSTED3:
                return tests.resources.pikamock.MockPikaBlockingConnection(params)
            else:
                raise ValueError("Unexpected RabbitMQ URL. Unit tests cannot work properly if the wrong test data is passed.")
        connectionmock.side_effect = make_mocked_connection

        # Make test rabbit:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()

        # Run code to be tested (during this, the connection is opened, i.e. the patch is needed)
        testrabbit.open_rabbit_connection()

        # Check result:
        # Connection created:
        conn = testrabbit._SynchronousRabbitConnector__connection
        self.assertIsInstance(conn, tests.resources.pikamock.MockPikaBlockingConnection)
        # Channel created:
        channel = testrabbit._SynchronousRabbitConnector__channel
        self.assertIsInstance(channel, tests.resources.pikamock.MockChannel)
        # Communcation established:
        established = testrabbit._SynchronousRabbitConnector__communication_established
        self.assertTrue(established)
        # Connected to fallback URL:
        expected_host = RABBIT_URL_TRUSTED3
        self.assertIn(conn.host, expected_host, 'Connected to: %s instead of %s' % (conn.host, expected_host))

    @unittest.skipIf(globalvar.QUICK_ONLY, '(this test is slow)')
    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_channel')
    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_open_rabbit_channel_fails(self, connectionmock, channelmock):

        print(self.slow_message)

        # Define replacement for mocked methods:
        # Connection succeeds, channel does not.
        def make_mocked_connection(params):
            return tests.resources.pikamock.MockPikaBlockingConnection(params)
        connectionmock.side_effect = make_mocked_connection
        channelmock.side_effect = pika.exceptions.ChannelClosed(reply_code=0, reply_text='Channel Closed')

        # Make test rabbit:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()

        # Run code to be tested (during this, the connection is opened, i.e. the patch is needed)
        with self.assertRaises(PIDServerException) as e:
            testrabbit.open_rabbit_connection()

        # Check result:
        # Connection created:
        conn = testrabbit._SynchronousRabbitConnector__connection
        self.assertIsInstance(conn, tests.resources.pikamock.MockPikaBlockingConnection)
        # Channel is None:
        channel = testrabbit._SynchronousRabbitConnector__channel
        self.assertIsNone(channel)
        # Communcation NOT established:
        established = testrabbit._SynchronousRabbitConnector__communication_established
        self.assertFalse(established)
        # Connected to fallback URL:
        expected_host = RABBIT_URL_TRUSTED3
        self.assertIn(conn.host, expected_host, 'Connected to: %s instead of %s' % (conn.host, expected_host))

    #
    # Close connection
    #

    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_close_rabbit_connection_ok(self, connectionmock):

        # Define replacement for mocked object:
        def make_mocked_connection(params):
            return tests.resources.pikamock.MockPikaBlockingConnection(params)
        connectionmock.side_effect = make_mocked_connection

        # Make test rabbit with open connection:
        #(during this, the connection is opened, i.e. the patch is needed)
        testrabbit = TESTHELPERS.get_synchronous_rabbit()
        testrabbit.open_rabbit_connection()

        # Run code to be tested:
        testrabbit.close_rabbit_connection()

        # Check result:
        # Connection should be closed:
        conn = testrabbit._SynchronousRabbitConnector__connection
        self.assertFalse(conn.is_open)

    #
    # Send message
    #

    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_send_message_ok(self, connectionmock):
        '''
        We expect the message to arrive in the MockChannel
        on the first try.
        '''

        # Test variables:
        key = 'mykey'
        msg = {"foo":"bar", "ROUTING_KEY":key}

        # Define replacement for mocked method:
        # A mocked connection is created.
        def make_mocked_connection(params):
            return tests.resources.pikamock.MockPikaBlockingConnection(params)
        connectionmock.side_effect = make_mocked_connection

        # Make test rabbit with connection:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()
        testrabbit.open_rabbit_connection()

        # Run code to be tested:
        testrabbit.send_message_to_queue(msg)

        # Check result:
        # Message was sent:
        channel = testrabbit._SynchronousRabbitConnector__channel
        self.assertEqual(channel.success_counter, 1, 'Message should succeed!')
        # Message was sent on first try:
        self.assertEqual(channel.publish_counter, 1,
            'Message was sent %i times!' % channel.publish_counter)
        # Message arrived in mock channel:
        self.assertIn(json.dumps(msg), channel.messages,
            'Message %s not found in channel\'s list: %s' % (msg, channel.messages))
        self.assertIn(key, channel.routing_keys,
            'Routing key %s not found in channel\'s list: %s' % (key, channel.routing_keys))

    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_send_message_fail(self, connectionmock):

        # Test variables:
        key = 'mykey'
        msg = {"foo":"bar", "ROUTING_KEY":key}

        # Define replacement for mocked method:
        # A mocked connection is created.
        def make_mocked_connection(params):
            return tests.resources.pikamock.MockPikaBlockingConnection(params)
        connectionmock.side_effect = make_mocked_connection

        # Make test rabbit with connection:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()
        testrabbit.open_rabbit_connection()

        # Make publish raise an UnroutableError
        channel = testrabbit._SynchronousRabbitConnector__channel
        channel.num_failures = float('inf') # always failure!

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.MessageNotDeliveredException) as e:
            testrabbit.send_message_to_queue(msg)

        # Check result:
        self.assertEqual(channel.success_counter, 0, 'Message should not succeed!')

    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_send_message_unroutable(self, connectionmock):

        # Test variables:
        key = 'mykey'
        msg = {"foo":"bar", "ROUTING_KEY":key}

        # Define replacement for mocked object:
        def make_mocked_connection(params):
            return tests.resources.pikamock.MockPikaBlockingConnection(params)
        connectionmock.side_effect = make_mocked_connection
        # Make basic_publish raise an error
        #publishmock.side_effect = pika.exceptions.UnroutableError('foo bar')

        # Make test rabbit with connection:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()
        testrabbit.open_rabbit_connection()

        # Make publish raise an UnroutableError
        channel = testrabbit._SynchronousRabbitConnector__channel
        channel.num_unroutables = float('inf') # always unroutable!

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.MessageNotDeliveredException) as e:
            testrabbit.send_message_to_queue(msg)

        # Check result:
        self.assertEqual(channel.success_counter, 0, 'Message should not succeed!')
        self.assertEqual(channel.publish_counter, 2, 'Message was tried %i times, not 2.' % channel.publish_counter)



    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_send_message_first_fails_then_ok(self, connectionmock):
        '''
        We make the channel return False (failure), then True (success).
        We check if the message arrived in the MockChannel.
        '''

        # Test variables:
        key = 'mykey'
        msg = {"foo":"bar", "ROUTING_KEY":key}

        # Define replacement for mocked object:
        def make_mocked_connection(params):
            return tests.resources.pikamock.MockPikaBlockingConnection(params)
        connectionmock.side_effect = make_mocked_connection

        # Make test rabbit with connection:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()
        testrabbit.open_rabbit_connection()

        # Make publish succeed only in 2nd try:
        channel = testrabbit._SynchronousRabbitConnector__channel
        channel.num_failures = 1

        # Run code to be tested:
        testrabbit.send_message_to_queue(msg)

        # Check result:
        # Message was sent:
        self.assertEqual(channel.success_counter, 1, 'Message should succeed!')
        # Message was sent on second try:
        self.assertEqual(channel.publish_counter, 2,
            'Message was sent %i times!' % channel.publish_counter)
        # Message arrived in mock channel:
        self.assertIn(json.dumps(msg), channel.messages,
            'Message %s not found in channel\'s list: %s' % (msg, channel.messages))
        self.assertIn(key, channel.routing_keys,
            'Routing key %s not found in channel\'s list: %s' % (key, channel.routing_keys))


    @mock.patch('esgfpid.rabbit.synchronous.SynchronousRabbitConnector._SynchronousRabbitConnector__make_connection')
    def test_send_message_no_connection(self, connectionmock):
        '''
        If there is no connection, it is built before publishing
        the message.
        We check if the message arrived in the MockChannel.
        '''

        # Test variables:
        key = 'mykey'
        msg = {"foo":"bar", "ROUTING_KEY":key}

        # Define replacement for mocked object:
        def make_mocked_connection(params):
            return tests.resources.pikamock.MockPikaBlockingConnection(params)
        connectionmock.side_effect = make_mocked_connection

        # Make test rabbit without connection:
        testrabbit = TESTHELPERS.get_synchronous_rabbit()

        # Run code to be tested:
        testrabbit.send_message_to_queue(msg)

        # Check result:
        # Connection is opened:
        conn = testrabbit._SynchronousRabbitConnector__connection
        self.assertTrue(conn.is_open)
        # Message was sent:
        channel = testrabbit._SynchronousRabbitConnector__channel
        self.assertEqual(channel.success_counter, 1, 'Message should succeed!')
        # Message was sent on first try:
        self.assertEqual(channel.publish_counter, 1,
            'Message was sent %i times!' % channel.publish_counter)
        # Message arrived in mock channel:
        self.assertIn(json.dumps(msg), channel.messages,
            'Message %s not found in channel\'s list: %s' % (msg, channel.messages))
        self.assertIn(key, channel.routing_keys,
            'Routing key %s not found in channel\'s list: %s' % (key, channel.routing_keys))


import datetime



def get_now_utc_as_formatted_string():
    now = get_now_utc()
    now_string = datetime.datetime.isoformat(now) # 2015-12-21T10:31:37.524825+00:00
    return now_string

def get_now_utc():
    ''' date in UTC, ISO format'''

    # Helper class for UTC time
    # Source: http://stackoverflow.com/questions/2331592/datetime-datetime-utcnow-why-no-tzinfo
    ZERO = datetime.timedelta(0)
    class UTC(datetime.tzinfo):
        """UTC"""
        def utcoffset(self, dt):
            return ZERO
        def tzname(self, dt):
            return "UTC"
        def dst(self, dt):
            return ZERO

    #now = datetime.datetime.now(timezone.utc) # Python 3.2
    now = datetime.datetime.now(UTC())
    return now
 


import esgfpid.utils
import esgfpid.exceptions
import logging
from . import utils as solrutils


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class FindFilesOfSameDatasetVersion(object):

    def __init__(self, solr_interactor):
        self.__solr_interactor = solr_interactor
        self.__error_messages = None

    def __reset_error_messages(self):
        self.__error_messages = []

    # General methods:

    def retrieve_file_handles_of_same_dataset(self, **args):
        '''
        :return: List of handles, or empty list. Should never return None.
        :raise: SolrSwitchedOff
        :raise SolrError: If both strategies to find file handles failed.
        '''
        mandatory_args = ['drs_id', 'version_number', 'data_node', 'prefix']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        self.__reset_error_messages()

        # Try plan A
        file_handles = None
        try:
            file_handles = self.__strategy1(args) # can raise SolrError or SolrSwitchedOff, but can't return None
        except esgfpid.exceptions.SolrError as e:
            self.__error_messages.append('Error during first query: '+repr(e))

        if file_handles is not None and len(file_handles)>0:
            LOGGER.debug('Retrieved file handles from solr in first query.')
            return file_handles

        # Try plan B
        try:
            file_handles = self.__strategy2(args) # can raise SolrError or SolrSwitchedOff, but can't return None
        except esgfpid.exceptions.SolrError as e:
            self.__error_messages.append('Error during second query: '+repr(e))
            msg = '/n'.join(self.__error_messages)
            raise esgfpid.exceptions.SolrError('Failure in both queries. Messages:\n'+msg)

        return file_handles

    def __strategy1(self, args):
        return self.__retrieve_file_handles_of_same_dataset_if_same_datanode(
            args['drs_id'],
            args['version_number'],
            args['data_node'],
            args['prefix'])

    def __strategy2(self, args):
        return self.__retrieve_file_handles_of_same_dataset_if_different_datanode(
            args['drs_id'],
            args['version_number'],
            args['prefix'])

    def __retrieve_file_handles_of_same_dataset_if_same_datanode(self, drs_id, version_number, data_node, prefix):
        dataset_id = self.__make_dataset_id_from_drsid_and_versnum(drs_id, version_number, data_node)
        response_json = self.__ask_solr_for_handles_of_files_with_same_dataset_id(dataset_id) # can raise SolrError or SolrSwitchedOff, but can't be None
        file_handles = solrutils.extract_file_handles_from_response_json(response_json, prefix) # can raise SolrReponseError
        if len(file_handles)==0:
            msg = 'First query returned an empty list.'
            LOGGER.debug(msg)
            self.__error_messages.append(msg)
        return file_handles

    def __make_dataset_id_from_drsid_and_versnum(self, drs_id, version_number, data_node):
        dataset_id = drs_id + '.v' + str(version_number) + '|' + data_node
        return dataset_id

    def __retrieve_file_handles_of_same_dataset_if_different_datanode(self, drs_id, version_number, prefix):
        response_json = self.__ask_solr_for_handles_of_files_with_same_drsid_and_version(drs_id, version_number) # can raise SolrError or SolrSwitchedOff, but can't be None
        file_handles = solrutils.extract_file_handles_from_response_json(response_json, prefix) # can raise SolrReponseError
        if len(file_handles)==0:
            self.__error_messages.append('Second query returned an empty list.')
        return file_handles

    # Querying solr for handles of files with the same dataset id:

    def __ask_solr_for_handles_of_files_with_same_dataset_id(self, dataset_id):
        LOGGER.debug('Asking solr for all handles or files with the dataset_id "%s".', dataset_id)
        query = self.__make_query_handles_of_files_with_same_dataset_id(dataset_id)
        LOGGER.debug('Query: %s', query)
        response_json = self.__solr_interactor.send_query(query) # can raise SolrError or SolrSwitchedOff, but can't be None
        return response_json

    def __make_query_handles_of_files_with_same_dataset_id(self, dataset_id):
        query_dict = self.__solr_interactor.make_solr_base_query()
        query_dict['type'] = 'File'
        query_dict['facets'] = 'tracking_id'
        query_dict['dataset_id'] = dataset_id
        return query_dict

    # Querying solr for the dataset_ids of all versions with the same drs_id:

    def __ask_solr_for_handles_of_files_with_same_drsid_and_version(self, drs_id, version_number):
        part_of_dataset_id = self.__make_half_dataset_id_from_drsid_and_versnum(drs_id, version_number)
        LOGGER.debug('Asking solr for all handles or files with the "half dataset_id" "%s".', part_of_dataset_id)
        query = self.__make_query_for_handles_of_files_with_same_drsid_and_version(part_of_dataset_id)
        LOGGER.debug('Query: %s', query)
        response_json = self.__solr_interactor.send_query(query) # can raise SolrError or SolrSwitchedOff, but can't be None
        return response_json

    def __make_half_dataset_id_from_drsid_and_versnum(self, drs_id, version_number):
        part_of_dataset_id = drs_id + '.v' + str(version_number) + '|'
        return part_of_dataset_id

    def __make_query_for_handles_of_files_with_same_drsid_and_version(self, part_of_dataset_id):
        query_dict = self.__solr_interactor.make_solr_base_query()
        query_dict['type'] = 'File'
        query_dict['facets'] = 'tracking_id'
        query_dict['query'] = 'dataset_id:'+part_of_dataset_id+'*'
        return query_dict




def replace_date_with_string(json_object, key='message_timestamp', replacement='anydate'):
    if key in json_object:
        json_object[key] = replacement

def is_json_same(expected, received):
    '''
    This function compares two JSON objects on equality, i.e.:
     - Do they have the same keys?
     - Do the shared keys have the same value?
    '''
    expected_keys = list(expected.keys())
    expected_keyscopy = expected_keys[:]
    received_keys = list(received.keys())
    for key in expected_keyscopy:
        if key not in received_keys:
            return False
        elif isinstance(expected[key], list):
            return True if sorted(expected[key]) == sorted(received[key]) else False
        elif expected[key] != received[key]:
            return False
        else:
            expected_keys.remove(key)
            received_keys.remove(key)
    if len(received_keys) > 0:
        return False

    return True

def compare_json_return_errormessage(expected, received, full=True):
    '''
    Basically the same as "is_json_same()", but it does not return
    on the first found error, and collects verbose information about
    errors on the way.

    :param full: Optional. Boolean. If True, appends both full objects
                 to the error message. Defaults to True.
    :return: Error message as string if an error was found.
             Otherwise, returns None.
    '''
    error_msg = ''
    error_occurred = False

    expected_keys = list(expected.keys())
    expected_keyscopy = expected_keys[:]
    received_keys = list(received.keys())
    for key in expected_keyscopy:
        if key not in received_keys:
            #error_msg += ('\nExpected key "'+key+'" not in received keys: '+str(received_keys))
            error_msg += ('\nExpected key "%s" not in received keys: %s' % (key, received_keys))
            error_occurred = True
        elif isinstance(expected[key], list):
            if sorted(expected[key]) != sorted(received[key]):
                error_msg += ('\nReceived for key "%s":\n\t%s (%s)\nnot same as expected\n\t%s (%s)' % (key, received[key], type(received[key]), expected[key], type(expected[key])))
                error_occurred = True
                expected_keys.remove(key)
                received_keys.remove(key)
        elif expected[key] != received[key]:
            #error_msg += ('\nReceived for key "'+str(key)+'":\n'+str(received[key])+'\nnot same as expected\n'+str(expected[key]))
            error_msg += ('\nReceived for key "%s":\n\t%s (%s)\nnot same as expected\n\t%s (%s)' % (key, received[key], type(received[key]), expected[key], type(expected[key])))
            error_occurred = True
            expected_keys.remove(key)
            received_keys.remove(key)
        else:
            expected_keys.remove(key)
            received_keys.remove(key)
    if len(received_keys) > 0:
        #error_msg += ('\nThese keys were received, but not expected: '+str(received_keys))
        error_msg += ('\nThese keys were received, but not expected: %s' % received_keys)
        error_occurred = True

    if error_occurred:
        if full:
            #error_msg += '\nExpected:\n'+str(expected)+'\nReceived:\n'+str(received)
            error_msg += ('\nExpected:\n\t%s\nReceived:\n\t%s' % (expected, received))
        return error_msg
    else:
        return None

#def compare_json(expected, received):
#    errormsg = compare_json_return_errormessage(expected, received)
#    if errormsg is None:
#        return True
#    else:
#        return False


def is_json_in_jsonlist(json_object, json_list):
    for item in json_list:
        is_same = compare_json(item, json_object)
        if is_same:
            return True
    return False

def remove_date_simple(jsonobject):
    try:
        jsonobject['publication_date'] = 'anydate'
    except KeyError:
        for message in jsonobject:
            message['publication_date'] = 'anydate'
    return jsonobject

def remove_date_recursive(jsonobject):
    return replace_value(jsonobject,'publication_date','anydate')

def replace_value(jsonobject, key_to_replace, new_value):
    if type(jsonobject) == dict:
        if jsonobject == {}:
            return jsonobject
        else:
            try:
                jsonobject[key_to_replace] = new_value
            except KeyError:
                pass
            for key in jsonobject:
                replace_value(jsonobject[key], key_to_replace, new_value)
            return jsonobject

def remove_comment(jsonobject):
    return remove_pair(jsonobject, 'comment')

def remove_pair(jsonobject, key_to_remove):
    if type(jsonobject) == dict:
        if jsonobject == {}:
            return jsonobject
        else:
            try:
                del jsonobject[key_to_remove]
            except KeyError:
                pass
            for key in jsonobject:
                remove_pair(jsonobject[key], key_to_remove)
            return jsonobject